from meddlr.modeling.blocks import conv_blocks, fuse_blocks
from meddlr.modeling.blocks.conv_blocks import (  # noqa: F401
    SimpleConvBlock2d,
    SimpleConvBlock3d,
    SimpleConvBlockNd,
)
from meddlr.modeling.blocks.fuse_blocks import (  # noqa: F401
    ConcatBlock2d,
    ConcatBlock3d,
    ConcatBlockNd,
    ResBlock3d,
    ResBlockNd,
)

__all__ = []
__all__.extend(conv_blocks.__all__)
__all__.extend(fuse_blocks.__all__)



from .config import CfgNode as CN

# -----------------------------------------------------------------------------
# Convention about Training / Test specific parameters
# -----------------------------------------------------------------------------
# Whenever an argument can be either used for training or for testing, the
# corresponding name will be post-fixed by a _TRAIN for a training parameter,
# or _TEST for a test-specific parameter.
# For example, the number of images during training will be
# IMAGES_PER_BATCH_TRAIN, while the number of images for testing will be
# IMAGES_PER_BATCH_TEST

# -----------------------------------------------------------------------------
# Base configurations
# -----------------------------------------------------------------------------
_MRI_RECON_TFM = CN()
_MRI_RECON_TFM.TRANSFORMS = ()
_MRI_RECON_TFM.SCHEDULER_P = CN(new_allowed=True)
_MRI_RECON_TFM.SCHEDULER_P.IGNORE = False
_MRI_RECON_TFM.AUG_SENSITIVITY_MAPS = True

# -----------------------------------------------------------------------------
# Config definition
# -----------------------------------------------------------------------------

_C = CN()
_C.VERSION = 1

_C.MODEL = CN()
_C.MODEL.DEVICE = "cuda"
_C.MODEL.META_ARCHITECTURE = "GeneralizedUnrolledCNN"
_C.MODEL.WEIGHTS = ""

# -----------------------------------------------------------------------------
# Unrolled model - TODO: Deprecate
# -----------------------------------------------------------------------------
_C.MODEL.UNROLLED = CN()
_C.MODEL.UNROLLED.BLOCK_ARCHITECTURE = "ResNet"
_C.MODEL.UNROLLED.NUM_UNROLLED_STEPS = 5
_C.MODEL.UNROLLED.NUM_RESBLOCKS = 2
_C.MODEL.UNROLLED.NUM_FEATURES = 256
_C.MODEL.UNROLLED.DROPOUT = 0.0
# Padding options. "" for now. TODO: add "circular"
_C.MODEL.UNROLLED.PADDING = ""
_C.MODEL.UNROLLED.FIX_STEP_SIZE = False
_C.MODEL.UNROLLED.SHARE_WEIGHTS = False
# Kernel size
_C.MODEL.UNROLLED.KERNEL_SIZE = (3,)
# Number of ESPIRiT maps
_C.MODEL.UNROLLED.NUM_EMAPS = 1

# Conv block parameters
_C.MODEL.UNROLLED.CONV_BLOCK = CN()
# Either "relu" or "leaky_relu"
_C.MODEL.UNROLLED.CONV_BLOCK.ACTIVATION = "relu"
# Either "none", "instance", or "batch"
_C.MODEL.UNROLLED.CONV_BLOCK.NORM = "none"
# Use affine on norm
_C.MODEL.UNROLLED.CONV_BLOCK.NORM_AFFINE = False
_C.MODEL.UNROLLED.CONV_BLOCK.ORDER = ("norm", "act", "drop", "conv")

_C.MODEL.RECON_LOSS = CN()
_C.MODEL.RECON_LOSS.NAME = "l1"
_C.MODEL.RECON_LOSS.RENORMALIZE_DATA = True
_C.MODEL.CONSISTENCY = CN()
_C.MODEL.CONSISTENCY.USE_LATENT = False
_C.MODEL.CONSISTENCY.USE_CONSISTENCY = True
_C.MODEL.CONSISTENCY.LATENT_LOSS_NAME = "mag_l1"
_C.MODEL.CONSISTENCY.NUM_LATENT_LAYERS = 1
_C.MODEL.CONSISTENCY.LOSS_NAME = "l1"
_C.MODEL.CONSISTENCY.LOSS_WEIGHT = 0.1
_C.MODEL.CONSISTENCY.LATENT_LOSS_WEIGHT = 0.1
# Consistency Augmentations
_C.MODEL.CONSISTENCY.AUG = CN()
_C.MODEL.CONSISTENCY.AUG.NOISE = CN()
_C.MODEL.CONSISTENCY.AUG.MOTION = CN()
# Noise standard deviation - 1,5,8 used for 3D FSE in Lustig paper.
_C.MODEL.CONSISTENCY.AUG.NOISE.STD_DEV = (1,)
_C.MODEL.CONSISTENCY.AUG.MOTION.RANGE = (0.2, 0.5)

# Noise scheduler
_C.MODEL.CONSISTENCY.AUG.NOISE.SCHEDULER = CN()
_C.MODEL.CONSISTENCY.AUG.NOISE.SCHEDULER.WARMUP_METHOD = ""
_C.MODEL.CONSISTENCY.AUG.NOISE.SCHEDULER.WARMUP_ITERS = 0
# Noise masking
_C.MODEL.CONSISTENCY.AUG.NOISE.MASK = CN()
_C.MODEL.CONSISTENCY.AUG.NOISE.MASK.RHO = 1.0
_C.MODEL.CONSISTENCY.AUG.MRI_RECON = _MRI_RECON_TFM.clone()

# Motion scheduler
_C.MODEL.CONSISTENCY.AUG.MOTION.SCHEDULER = CN()
_C.MODEL.CONSISTENCY.AUG.MOTION.SCHEDULER.WARMUP_METHOD = ""
_C.MODEL.CONSISTENCY.AUG.MOTION.SCHEDULER.WARMUP_ITERS = 0

# Image normalization options.
# See ss_recon/data/transforms/transform.py for more details.
_C.MODEL.NORMALIZER = CN()
_C.MODEL.NORMALIZER.NAME = "TopMagnitudeNormalizer"
# Keywords to normalize
# See ss_recon/data/transforms/transform.py for more details
_C.MODEL.NORMALIZER.KEYWORDS = ()

_C.MODEL.SEG = CN()
_C.MODEL.SEG.CLASSES = ()
_C.MODEL.SEG.INCLUDE_BACKGROUND = False
_C.MODEL.SEG.ACTIVATION = "sigmoid"

# -----------------------------------------------------------------------------
# UNET model - shared by UNetModel and GeneralizedUNet modules
# -----------------------------------------------------------------------------
_C.MODEL.UNET = CN()
_C.MODEL.UNET.OUT_CHANNELS = 2
_C.MODEL.UNET.IN_CHANNELS = 2
_C.MODEL.UNET.CHANNELS = 32
_C.MODEL.UNET.NUM_POOL_LAYERS = 4
_C.MODEL.UNET.DROPOUT = 0.0
_C.MODEL.UNET.NORMALIZE = False
# BLOCK_ORDER only applies to GeneralizedUNet
_C.MODEL.UNET.BLOCK_ORDER = ("conv", "relu", "conv", "relu", "batchnorm", "dropout")

# -----------------------------------------------------------------------------
# Denoising model
# -----------------------------------------------------------------------------
_C.MODEL.DENOISING = CN()
_C.MODEL.DENOISING.META_ARCHITECTURE = "GeneralizedUnrolledCNN"
_C.MODEL.DENOISING.NOISE = CN()
# Noise standard deviation to use for augmentations.
_C.MODEL.DENOISING.NOISE.STD_DEV = (1,)
# When fully sampled kspace is available, if True, perform denoising on the
# fully sampled kspace. If False, denoising will be performed on the
# randomly generated undersampled kspace.
_C.MODEL.DENOISING.NOISE.USE_FULLY_SAMPLED_TARGET = True
# Same as above, but at eval time (e.g. validation).
# Defaults to MODEL.DENOISING.NOISE.USE_FULLY_SAMPLED_TARGET
_C.MODEL.DENOISING.NOISE.USE_FULLY_SAMPLED_TARGET_EVAL = None

# -----------------------------------------------------------------------------
# Compressed Sensing (CS) model
# -----------------------------------------------------------------------------
_C.MODEL.CS = CN()
_C.MODEL.CS.REGULARIZATION = 0.005
_C.MODEL.CS.MAX_ITER = 200

# -----------------------------------------------------------------------------
# Noise2Recon (N2R) model
# -----------------------------------------------------------------------------
_C.MODEL.N2R = CN()
_C.MODEL.N2R.META_ARCHITECTURE = "GeneralizedUnrolledCNN"
_C.MODEL.N2R.USE_SUPERVISED_CONSISTENCY = False
# Use MODEL.CONSISTENCY.AUG.NOISE.STD_DEV to control noise parameters.

# -----------------------------------------------------------------------------
# Motion2Recon (M2R) model
# -----------------------------------------------------------------------------
_C.MODEL.M2R = CN()
_C.MODEL.M2R.META_ARCHITECTURE = "GeneralizedUnrolledCNN"
_C.MODEL.M2R.USE_SUPERVISED_CONSISTENCY = False
# Use MODEL.CONSISTENCY.AUG.MOTION_RANGE to control motion parameters.

# -----------------------------------------------------------------------------
# NoiseMotion2Recon (NM2R) model
# -----------------------------------------------------------------------------
_C.MODEL.NM2R = CN()
_C.MODEL.NM2R.META_ARCHITECTURE = "GeneralizedUnrolledCNN"
_C.MODEL.NM2R.USE_SUPERVISED_CONSISTENCY = False
# Use MODEL.CONSISTENCY.AUG.MOTION_RANGE to control motion parameters.

# -----------------------------------------------------------------------------
# Aug2Recon (A2R) / VORTEX model
# -----------------------------------------------------------------------------
_C.MODEL.A2R = CN()
_C.MODEL.A2R.META_ARCHITECTURE = "GeneralizedUnrolledCNN"
_C.MODEL.A2R.USE_SUPERVISED_CONSISTENCY = False

# -----------------------------------------------------------------------------
# SSDU model
# -----------------------------------------------------------------------------
_C.MODEL.SSDU = CN()
_C.MODEL.SSDU.META_ARCHITECTURE = "GeneralizedUnrolledCNN"
_C.MODEL.SSDU.MASKER = CN()
_C.MODEL.SSDU.MASKER.PARAMS = CN(new_allowed=True)

# -----------------------------------------------------------------------------
# Dataset
# -----------------------------------------------------------------------------
_C.DATASETS = CN()
# List of the dataset names for training. Must be registered in DatasetCatalog
_C.DATASETS.TRAIN = ()
# List of the dataset names for validation. Must be registered in DatasetCatalog
_C.DATASETS.VAL = ()
# List of the dataset names for testing. Must be registered in DatasetCatalog
_C.DATASETS.TEST = ()


# -----------------------------------------------------------------------------
# DataLoader
# -----------------------------------------------------------------------------
_C.DATALOADER = CN()
# Number of data loading threads
_C.DATALOADER.NUM_WORKERS = 4
# Number of batches to prefetch per worker
_C.DATALOADER.PREFETCH_FACTOR = 2
# If True, the dataloader will drop the last batch.
_C.DATALOADER.DROP_LAST = True
# Subsample training data to simulate data limited scenarios.
_C.DATALOADER.SUBSAMPLE_TRAIN = CN()
# Number of training examples to retain. All others will be force dropped
# meaning they will not be used for any training, even w/ semi-supervised N2R
# framework
_C.DATALOADER.SUBSAMPLE_TRAIN.NUM_TOTAL = -1
# Number of training examples to retain based on metadata properties.
# e.g. (("acquisition", {"AXT2": 15, "AXT1": 10}),) will only keep
# 15 AXT2 scans and 10 AXT1 scans where AXT2 and AXT1 are categories in the
# "acquisition" metadata field.
_C.DATALOADER.SUBSAMPLE_TRAIN.NUM_TOTAL_BY_GROUP = ()
# Number of scans out of total to undersample. If NUM_TOTAL is not -1, must be
# less than NUM_TOTAL.
_C.DATALOADER.SUBSAMPLE_TRAIN.NUM_UNDERSAMPLED = 0
# Number of scans with ground truth for validation.
_C.DATALOADER.SUBSAMPLE_TRAIN.NUM_VAL = -1
# Number of validation examples to retain based on metadata properties.
_C.DATALOADER.SUBSAMPLE_TRAIN.NUM_VAL_BY_GROUP = ()
# Seed for shuffling data. Should always be deterministic
_C.DATALOADER.SUBSAMPLE_TRAIN.SEED = 1000
# Options: "" (defaults to random sampling), "AlternatingSampler"
_C.DATALOADER.SAMPLER_TRAIN = ""
# AlternatingSampler config parameters.
_C.DATALOADER.ALT_SAMPLER = CN()
_C.DATALOADER.ALT_SAMPLER.PERIOD_SUPERVISED = 1
_C.DATALOADER.ALT_SAMPLER.PERIOD_UNSUPERVISED = 1
# GroupSampler config parameters.
_C.DATALOADER.GROUP_SAMPLER = CN()
_C.DATALOADER.GROUP_SAMPLER.BATCH_BY = ()
_C.DATALOADER.GROUP_SAMPLER.AS_BATCH_SAMPLER = False
# Paired tuple of data keys and H5DF keys. Empty tuple will result in default keys being used.
# e.g. (("target", "espirit_recon"), ("maps", "espirit_maps"))
# See data/slice_data.py for more information.
_C.DATALOADER.DATA_KEYS = ()

_C.DATALOADER.FILTER = CN()
# Paired tuple of key and values for filtering. Multiple values should be specified as tuple.
# e.g. (("num_slices", 30),) will only keep data with number of slices = 30
# Field must appear in all dataset dicts.
# Training, validation, and testing data is filtered by this field.
_C.DATALOADER.FILTER.BY = ()


# -----------------------------------------------------------------------------
# Augmentations/Transforms
# -----------------------------------------------------------------------------
_C.AUG_TRAIN = CN()
_C.AUG_TRAIN.UNDERSAMPLE = CN()
_C.AUG_TRAIN.UNDERSAMPLE.NAME = "PoissonDiskMaskFunc"
_C.AUG_TRAIN.UNDERSAMPLE.ACCELERATIONS = (6,)
_C.AUG_TRAIN.UNDERSAMPLE.CALIBRATION_SIZE = 20
_C.AUG_TRAIN.UNDERSAMPLE.CENTER_FRACTIONS = ()
_C.AUG_TRAIN.UNDERSAMPLE.MAX_ATTEMPTS = 30
# Augment with noise. See MODEL.CONSISTENCY.AUG.NOISE for parameters.
_C.AUG_TRAIN.USE_NOISE = False
_C.AUG_TRAIN.USE_MOTION = False
_C.AUG_TRAIN.NOISE_P = 0.2
_C.AUG_TRAIN.MOTION_P = 0.2
# Augmentations using MRIReconAugmentor pipeline.
# If this is specified, it will be used in place of MOTION and NOISE
# augmentation parameters above. If you would like to use those parameters
# set AUG_TRAIN.MRI_RECON.TRANSFORMS=().
# This augmentation pipeline will automatically be ignored at test time (for now).
_C.AUG_TRAIN.MRI_RECON = _MRI_RECON_TFM.clone()

_C.AUG_TEST = CN()
_C.AUG_TEST.UNDERSAMPLE = CN()
_C.AUG_TEST.UNDERSAMPLE.ACCELERATIONS = (6,)

# ---------------------------------------------------------------------------- #
# Solver
# ---------------------------------------------------------------------------- #
_C.SOLVER = CN()

_C.SOLVER.OPTIMIZER = "Adam"

# See ss_recon/solver/build.py for LR scheduler options
_C.SOLVER.LR_SCHEDULER_NAME = "WarmupMultiStepLR"

_C.SOLVER.MAX_ITER = 20

_C.SOLVER.BASE_LR = 1e-4

_C.SOLVER.MOMENTUM = 0.9

_C.SOLVER.WEIGHT_DECAY = 0.0001
# The weight decay that's applied to parameters of normalization layers
# (typically the affine transformation)
_C.SOLVER.WEIGHT_DECAY_NORM = 0.0

_C.SOLVER.GAMMA = 0.1
# The iteration number to decrease learning rate by GAMMA.
_C.SOLVER.STEPS = (30000,)

_C.SOLVER.WARMUP_FACTOR = 1.0 / 1000
_C.SOLVER.WARMUP_ITERS = 1000
_C.SOLVER.WARMUP_METHOD = "linear"

_C.SOLVER.GRAD_ACCUM_ITERS = 1

# Save a checkpoint after every this number of iterations
_C.SOLVER.CHECKPOINT_PERIOD = 1

# Number of images per batch across all machines.
# If we have 16 GPUs and IMS_PER_BATCH = 32,
# each GPU will see 2 images per batch.
_C.SOLVER.TRAIN_BATCH_SIZE = 16
_C.SOLVER.TEST_BATCH_SIZE = 16

_C.SOLVER.BIAS_LR_FACTOR = 1.0
_C.SOLVER.WEIGHT_DECAY_BIAS = _C.SOLVER.WEIGHT_DECAY

# ---------------------------------------------------------------------------- #
# Specific test options
# ---------------------------------------------------------------------------- #
_C.TEST = CN()
# The period over which to evaluate the model during training.
# Set to 0 to disable.
_C.TEST.EVAL_PERIOD = 1
# For end-to-end tests to verify the expected accuracy.
# Each item is [task, metric, value, tolerance]
# e.g.: [['bbox', 'AP', 38.5, 0.2]]
_C.TEST.EXPECTED_RESULTS = []
# Validate with test-like functionality.
# If True, undersampling masks for every validation scan will be fixed
# given an acceleration.
_C.TEST.VAL_AS_TEST = True
# Validation metrics.
_C.TEST.VAL_METRICS = CN()
_C.TEST.VAL_METRICS.RECON = ()
# The period over which to flush evaluation results.
# Specify if you are getting OOM errors during evaluation.
# Set to 0 to disable.
_C.TEST.FLUSH_PERIOD = 0

# ---------------------------------------------------------------------------- #
# Misc options
# ---------------------------------------------------------------------------- #
# Directory where output files are written
_C.OUTPUT_DIR = ""
# Set seed to negative to fully randomize everything.
# Set seed to positive to use a fixed seed. Note that a fixed seed does not
# guarantee fully deterministic behavior.
_C.SEED = -1
# The period (in terms of steps) for minibatch visualization at train time.
# Set to 0 to disable. Currently not functional.
# This will always be in number of iterations.
_C.VIS_PERIOD = 0
# The scale when referring to time generally in the config.
# Note that there are certain fields, which are explicitly at the iteration
# scale (e.g. GRAD_ACCUM_ITERS).
# Either "epoch" or "iter"
_C.TIME_SCALE = "epoch"
_C.CUDNN_BENCHMARK = False

# ---------------------------------------------------------------------------- #
# Config Description
# ---------------------------------------------------------------------------- #
_C.DESCRIPTION = CN()
# Brief description about config
_C.DESCRIPTION.BRIEF = ""
# The entity (team/individual) account for logging to Weights & Biases
_C.DESCRIPTION.ENTITY_NAME = "ss_recon"
# Project name for logging to Weights & Biases
_C.DESCRIPTION.PROJECT_NAME = "ss_recon"
# Experiment name for logging to Weights & Biases
_C.DESCRIPTION.EXP_NAME = ""
# Tags associated with experiment.
# e.g. "fastmri_knee_mc" for fastMRI dataset; "unrolled" for using unrolled network; etc.
_C.DESCRIPTION.TAGS = ()



import warnings
from typing import Any, Callable, TypeVar

TCallable = TypeVar("TCallable", bound=Callable[..., Any])


def deprecated(
    reason=None, vdeprecated=None, vremove=None, replacement=None
) -> Callable[[TCallable], TCallable]:
    def fn(func: TCallable) -> TCallable:
        msg = _get_deprecated_msg(func, reason, vdeprecated, vremove, replacement)
        warnings.warn(msg, DeprecationWarning)
        return func

    return fn


def _get_deprecated_msg(wrapped, reason, vdeprecated, vremoved, replacement=None):
    fmt = "{name} is deprecated"
    if vdeprecated:
        fmt += " since v{vdeprecated}"
    if vremoved:
        fmt += " and will be removed in v{vremoved}"
    fmt += "."

    if reason:
        fmt += " ({reason})"
    if replacement:
        fmt += " -- Use meddlr.{replacement} instead."

    return fmt.format(
        name=wrapped.__name__,
        reason=reason or "",
        vdeprecated=vdeprecated or "",
        vremoved=vremoved or "",
        replacement=replacement or "",
    )



from typing import Any, Dict, Tuple, Union

import torch.nn as nn

from meddlr.modeling.layers.build import get_layer_kind, get_layer_type

__all__ = [
    "SimpleConvBlockNd",
    "SimpleConvBlock2d",
    "SimpleConvBlock3d",
]


class SimpleConvBlockNd(nn.Sequential):
    """A convolutional block supporting normalization, conv, activation, and dropout.

    The block implements same padding and convolution stride of 1. The first conv layer will
    change the number of channels from `in_channels` to `out_channels`.

    The order of layers can be specified by certain keywords:
        * "conv": Convolution layer
        * "convws": Convolution + Weight Standardization layer
        * "batchnorm"/"bn": Batch Normalization
        * "instancenorm": Instance Normalization
        * "groupnorm": Group Normalization
        * "relu": ReLU
        * "dropout": Dropout

    Args:
        in_channels (int): Number of channels in the input.
        out_channels (int): Number of channels in the output.
        kernel_size (`int(s)`): Convolution kernel size.
        dimension (int): Integer specifying the dimension of convolution.
        dropout (float, optional): Dropout probability.
        order (:obj:`str(s)`, optional): Order of layers in the convolution block. Note layers
            can be repeated.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, ...]],
        dimension: int,
        stride: Union[int, Tuple[int, ...]] = 1,
        dropout: float = 0.0,
        order: Tuple[Union[str, Tuple[str, Dict]], ...] = ("conv", "batchnorm", "relu", "dropout"),
        padding: Union[str, int, Tuple[int, ...]] = "same",
    ):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.dropout_prob = dropout

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size,) * dimension
        else:
            assert len(kernel_size) == dimension

        if not padding:
            padding = 0
        if padding == "same":
            if not all(k % 2 == 1 for k in kernel_size):
                raise ValueError(f"Kernel sizes must be odd - got {kernel_size}")
            padding = tuple(k // 2 for k in kernel_size)
        elif not isinstance(padding, int) and not isinstance(padding, Tuple):  # TODO: Improve check
            raise ValueError(f"Invalid value for padding '{padding}'")

        names = [x if isinstance(x, str) else x[0] for x in order]
        layer_classes = [get_layer_type(layer_name, dimension) for layer_name in names]
        layer_kinds = [get_layer_kind(x) for x in layer_classes]

        layers = []
        running_num_channels = in_channels
        for idx, (name, layer_cls, kind) in enumerate(zip(order, layer_classes, layer_kinds)):
            curr_layer = order[idx]
            lyr_kwargs: Dict[str, Any] = curr_layer[1] if isinstance(curr_layer[1], dict) else {}
            if kind == "conv":
                layer = layer_cls(
                    running_num_channels,
                    out_channels,
                    kernel_size=kernel_size,
                    padding=padding,
                    stride=stride,
                    **lyr_kwargs,
                )
                running_num_channels = out_channels
            elif kind == "norm":
                layer = layer_cls(running_num_channels, **lyr_kwargs)
            elif kind == "dropout":
                layer = layer_cls(dropout, **lyr_kwargs)
            elif kind == "act":
                layer = layer_cls(**lyr_kwargs)
            else:
                raise ValueError(f"Layer {name} (kind: {kind}) not supported")
            layers.append(layer)

        # Define forward pass
        super().__init__(*layers)


class SimpleConvBlock2d(SimpleConvBlockNd):
    """2D implementation of :class:`SimpleConvBlockNd`.

    Args:
        in_channels (int): Number of channels in the input.
        out_channels (int): Number of channels in the output.
        kernel_size (`int(s)`): Convolution kernel size.
        dropout (float, optional): Dropout probability.
        order (:obj:`str(s)`, optional): Order of layers in the convolution block. Note layers
            can be repeated.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride: Union[int, Tuple[int, int]] = 1,
        dropout: float = 0.0,
        order: Tuple[str, ...] = ("conv", "batchnorm", "relu", "dropout"),
        padding: Union[str, int, Tuple[int, ...]] = "same",
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            dimension=2,
            stride=stride,
            dropout=dropout,
            order=order,
            padding=padding,
        )


class SimpleConvBlock3d(SimpleConvBlockNd):
    """3D implementation of :class:`SimpleConvBlockNd`.

    Args:
        in_channels (int): Number of channels in the input.
        out_channels (int): Number of channels in the output.
        kernel_size (`int(s)`): Convolution kernel size.
        dropout (float, optional): Dropout probability.
        order (:obj:`str(s)`, optional): Order of layers in the convolution block. Note layers
            can be repeated.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int, int]],
        stride: Union[int, Tuple[int, int]] = 1,
        dropout: float = 0.0,
        order: Tuple[str, ...] = ("conv", "batchnorm", "relu", "dropout"),
        padding: Union[str, int, Tuple[int, ...]] = "same",
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            dimension=3,
            stride=stride,
            dropout=dropout,
            order=order,
            padding=padding,
        )



"""Training loop.

Adapted from
https://github.com/facebookresearch/detectron2
"""

import logging
import time
import weakref

import torch

from meddlr.utils.events import EventStorage
from meddlr.utils.general import flatten_dict

__all__ = ["HookBase", "TrainerBase", "SimpleTrainer"]


class HookBase:
    """
    Base class for hooks that can be registered with :class:`TrainerBase`.

    Each hook can implement 4 methods. The way they are called is demonstrated
    in the following snippet:

    .. code-block:: python

        hook.before_train()
        for iter in range(start_iter, max_iter):
            hook.before_step()
            trainer.run_step()
            hook.after_step()
        hook.after_train()

    Notes:
        1. In the hook method, users can access `self.trainer` to access more
           properties about the context (e.g., current iteration).

        2. A hook that does something in :meth:`before_step` can often be
           implemented equivalently in :meth:`after_step`.
           If the hook takes non-trivial time, it is strongly recommended to
           implement the hook in :meth:`after_step` instead of
           :meth:`before_step`.
           The convention is that :meth:`before_step` should only take
           negligible time.

           Following this convention will allow hooks that do care about the
           difference
           between :meth:`before_step` and :meth:`after_step` (e.g., timer) to
           function properly.

    Attributes:
        trainer: A weak reference to the trainer object. Set by the trainer
            when the hook is registered.
    """

    def before_train(self):
        """
        Called before the first iteration.
        """
        pass

    def after_train(self):
        """
        Called after the last iteration.
        """
        pass

    def before_step(self):
        """
        Called before each iteration.
        """
        pass

    def after_step(self):
        """
        Called after each iteration.
        """
        pass


class TrainerBase:
    """
    Base class for iterative trainer with hooks.

    The only assumption we made here is: the training runs in a loop.
    A subclass can implement what the loop is.
    We made no assumptions about the existence of dataloader, optimizer, model,
    etc.

    Attributes:
        iter(int): the current iteration.

        start_iter(int): The iteration to start with.
            By convention the minimum possible value is 0.

        max_iter(int): The iteration to end training.

        storage(EventStorage): An EventStorage that's opened during the course
            of training.
    """

    def __init__(self):
        self._hooks = []

    def register_hooks(self, hooks):
        """
        Register hooks to the trainer. The hooks are executed in the order
        they are registered.

        Args:
            hooks (list[Optional[HookBase]]): list of hooks
        """
        hooks = [h for h in hooks if h is not None]
        for h in hooks:
            assert isinstance(h, HookBase)
            # To avoid circular reference, hooks and
            # trainer cannot own each other.
            # This normally does not matter, but will cause memory leak if the
            # involved objects contain __del__:
            # See http://engineering.hearsaysocial.com/2013/06/16/circular-references-in-python/  # noqa
            h.trainer = weakref.proxy(self)
        self._hooks.extend(hooks)

    def train(self, start_iter: int, max_iter: int):
        """
        Args:
            start_iter, max_iter (int): See docs above
        """
        logger = logging.getLogger(__name__)
        logger.info("Starting training from iteration {}".format(start_iter))

        self.iter = self.start_iter = start_iter
        self.max_iter = max_iter

        with EventStorage(start_iter) as self.storage:
            try:
                self.before_train()
                for self.iter in range(start_iter, max_iter):
                    self.before_step()
                    self.run_step()
                    self.after_step()
            except Exception as e:
                logger.error(e)
                raise e
            finally:
                self.after_train()

    def before_train(self):
        for h in self._hooks:
            h.before_train()

    def after_train(self):
        for h in self._hooks:
            h.after_train()

    def before_step(self):
        for h in self._hooks:
            h.before_step()

    def after_step(self):
        for h in self._hooks:
            h.after_step()
        # this guarantees, that in each hook's after_step,
        # storage.iter == trainer.iter
        self.storage.step()

    def run_step(self):
        raise NotImplementedError


class SimpleTrainer(TrainerBase):
    """
    A simple trainer for the most common type of task:
    single-cost single-optimizer single-data-source iterative optimization.
    It assumes that every step, you:

    1. Compute the loss with a data from the data_loader.
    2. Compute the gradients with the above loss.
    3. Update the model with the optimizer.

    If you want to do anything fancier than this,
    either subclass TrainerBase and implement your own `run_step`,
    or write your own training loop.
    """

    def __init__(self, model, data_loader, optimizer, loss_computer, metrics_computer=None):
        """
        Args:
            model: a torch Module. Takes a data from data_loader and returns a
                dict of losses.
            data_loader: an iterable. Contains data to be used to call model.
            optimizer: a torch optimizer.
            loss_computer: A callable that returns a dict of losses.
                All terms must have the word "loss" to be a valid loss.
            metrics_computer: A callable that returns a dict of metrics.
                Will not be used for loss.
        """
        super().__init__()

        """
        We set the model to training mode in the trainer.
        However it's valid to train a model that's in eval mode.
        If you want your model (or a submodule of it) to behave
        like evaluation during training, you can overwrite its train() method.
        """
        model.train()

        self.model = model
        self.data_loader = data_loader
        self._data_loader_iter = None
        self.optimizer = optimizer
        self.loss_computer = loss_computer
        self.metrics_computer = metrics_computer

    def before_train(self):
        out = super().before_train()
        self._data_loader_iter = iter(self.data_loader)
        return out

    def run_step(self):
        """
        Implement the standard training logic described above.
        """
        assert self.model.training, "[SimpleTrainer] model was changed to eval mode!"
        start = time.perf_counter()
        """
        If your want to do something with the data, you can wrap the dataloader.
        """
        try:
            inputs = next(self._data_loader_iter)
        except StopIteration:
            # Epoch has ended, reinitialize the iterator.
            self._data_loader_iter = iter(self.data_loader)
            inputs = next(self._data_loader_iter)

        data_time = time.perf_counter() - start

        """
        If your want to do something with the losses, you can wrap the model.
        """
        output_dict = self.model(inputs)
        output_dict.update({k: inputs[k] for k in ["mean", "std", "norm"] if k in inputs})
        loss_dict = {k: v for k, v in output_dict.items() if "loss" in k}
        loss_dict.update(self.loss_computer(inputs, output_dict))

        # losses = sum(v for k, v in loss_dict.items() if "loss" in k)
        losses = loss_dict["loss"]
        self._detect_anomaly(losses, loss_dict)

        metrics_dict = loss_dict
        metrics_dict.update(
            flatten_dict(
                {k: v for k, v in inputs.get("metrics", {}).items() if k not in metrics_dict}
            )
        )
        metrics_dict["data_time"] = data_time
        metrics_dict["total_loss"] = losses
        metrics_dict.update(self.metrics_computer(output_dict) if self.metrics_computer else {})
        self._write_metrics(metrics_dict)

        """
        If you need accumulate gradients or something similar, you can
        wrap the optimizer with your custom `zero_grad()` method.
        """
        self.optimizer.zero_grad()
        losses.backward()

        """
        If you need gradient clipping/scaling or other processing, you can
        wrap the optimizer with your custom `step()` method.
        """
        self.optimizer.step()

    def _detect_anomaly(self, losses, loss_dict):
        if not torch.isfinite(losses).all():
            raise FloatingPointError(
                "Loss became infinite or NaN at iteration={}!\n"
                "loss_dict = {}".format(self.iter, loss_dict)
            )

    def _write_metrics(self, metrics_dict: dict):
        """
        Args:
            metrics_dict (dict): dict of scalar metrics
        """
        metrics_dict = {
            k: torch.mean(v.detach().cpu()).item() if isinstance(v, torch.Tensor) else float(v)
            for k, v in metrics_dict.items()
        }

        if len(metrics_dict) > 1:
            self.storage.put_scalars(**metrics_dict)



import logging
import os
import shutil
import weakref
from abc import ABC, abstractmethod
from typing import Any, Optional

from iopath.common.file_io import PathHandler, PathManager

import meddlr as mr
from meddlr.utils import env
from meddlr.utils.cluster import Cluster

try:
    import gdown

    _GDOWN_AVAILABLE = True
except ImportError:
    _GDOWN_AVAILABLE = False

_REPO_DIR = os.path.join(os.path.dirname(__file__), "../..")
_LOGGER = logging.getLogger(__name__)

__all__ = ["GeneralPathHandler", "GithubHandler"]


class GeneralPathHandler(PathHandler, ABC):
    PREFIX = ""

    def _get_supported_prefixes(self):
        return [self.PREFIX]

    def _get_local_path(self, path: str, **kwargs):
        name = path[len(self.PREFIX) :]
        return os.path.join(self._root_dir(), name)

    def _open(self, path, mode="r", **kwargs):
        return env.get_path_manager().open(self._get_local_path(path), mode, **kwargs)

    def _mkdirs(self, path: str, **kwargs):
        os.makedirs(self._get_local_path(path), exist_ok=True)

    @abstractmethod
    def _root_dir(self) -> str:
        pass


class DataHandler(GeneralPathHandler):
    PREFIX = "data://"

    def _root_dir(self):
        return Cluster.working_cluster().get_path("data_dir")


class ResultsHandler(GeneralPathHandler):
    PREFIX = "results://"

    def _root_dir(self):
        return Cluster.working_cluster().get_path("results_dir")


class CacheHandler(GeneralPathHandler):
    PREFIX = "cache://"

    def _root_dir(self):
        return Cluster.working_cluster().get_path("cache_dir")


class GithubHandler(GeneralPathHandler):
    """Handler to download, cache, and match Github files to local folder structure.

    Often, it may be useful to reference files from the meddlr Github repository.
    This class (when used with :cls:`iopath.common.file_io.PathHandler`) fetches
    and locally caches files from the Meddlr Github repository.

    This class fetches files from the most recent tagged version in Meddlr.

    Attributes:
        github_url (str): The base URL for the Github repository.
        default_branch_or_tag (str): The default branch or tag to download files from.

    Examples:
        >>> gh = GithubHandler(github_url="https://github.com/ad12/meddlr")
        >>> gh.get_local_path("github://annotations")
    """

    PREFIX = "github://"

    def __init__(
        self,
        github_url: str = None,
        default_branch_or_tag: str = "main",
        default_cache_path: str = None,
        async_executor=None,
    ) -> None:
        if github_url is None:
            github_url = env.get_github_url()
        self.github_url = github_url
        self.default_branch_or_tag = default_branch_or_tag
        self.default_cache_path = default_cache_path
        super().__init__(async_executor=async_executor)

    def _root_dir(self):
        return None

    def _get_default_branch_or_tag(self):
        return f"v{mr.__version__}"

    def _get_default_cache_path(self):
        return os.path.join(
            env.get_path_manager().get_local_path(Cluster.working_cluster().cache_dir),
            f"github-repo/v{mr.__version__}",
        )

    def download(self, cache_dir, branch_or_tag: str = None, force: bool = False):
        if branch_or_tag is None:
            branch_or_tag = self.default_branch_or_tag
        if branch_or_tag is None:
            branch_or_tag = self._get_default_branch_or_tag()
        return download_github_repository(
            url=self.github_url, branch_or_tag=branch_or_tag, cache_path=cache_dir, force=force
        )

    def _get_local_path(
        self,
        path: str,
        branch_or_tag: str = None,
        force: bool = False,
        cache_dir: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        """Get local path to GitHub repository for the current meddlr version.

        This implementation will only download the GitHub repository if it does
        not exist. To force a download, set ``force=True``.

        Args:
            path (str): The relative file path in the GitHub repository.
                Must start with ``'github://'``.
            branch_or_tag (str): The Github branch or tag to download from.
            force (bool, optional): If ``True``, force a download of the Github
                repository. Defaults to ``False``.
            cache_dir (str, optional): The local cache directory to use.

        Returns:
            str: The local path to the file/directory.
        """
        path = path[len(self.PREFIX) :]
        self._check_kwargs(kwargs)
        if cache_dir is None:
            cache_dir = (
                self.default_cache_path
                if self.default_cache_path
                else self._get_default_cache_path()
            )

        if force or not os.path.exists(cache_dir) or not os.listdir(cache_dir):
            logger = logging.getLogger(__name__)
            logger.info(
                "Downloading github repository (tag/branch {}) from {}...".format(
                    mr.__version__, self.github_url
                )
            )
            self.download(cache_dir, force=force, branch_or_tag=branch_or_tag)
            logger.info("Repository cached to {}".format(cache_dir))

        expected_path = os.path.join(str(cache_dir), path)
        if not os.path.exists(expected_path):
            raise FileNotFoundError(
                "Could not find {} in local Github cache: {}".format(path, cache_dir)
            )
        return expected_path


class AnnotationsHandler(GeneralPathHandler):
    PREFIX = "ann://"

    def _get_cache_path(self):
        return "github://annotations"

    def _root_dir(self):
        local_path = os.path.abspath(os.path.join(_REPO_DIR, "annotations"))
        if os.path.isdir(local_path):
            return local_path

        cache_path = env.get_path_manager().get_local_path(self._get_cache_path())
        if not os.path.isdir(cache_path):
            self.download()
        return cache_path

    def download(self):
        """Downloads annotations from meddlr github.

        Note:
            This method downloads the full repository of the current
            meddlr version and then copies the annotation files to
            the respective directory. This may not be very efficient,
            but it ensures that the annotation download process is
            less onerous on the user.
        """
        _LOGGER.info("Downloading annotations...")
        repo_path = download_github_repository()
        cache_path = env.get_path_manager().get_local_path(self._get_cache_path())
        os.makedirs(cache_path, exist_ok=True)
        retval = os.system(f"cp -r {repo_path}/annotations {cache_path}/")
        if retval != 0:
            raise RuntimeError("Could not download annotations.")


class GoogleDriveHandler(GeneralPathHandler):
    """Handler to download, cache, and match Google drive files to local folder structure.

    Publicly available files on google drive can be downloaded using this handler.
    By default, files are cached to "~/.cache/gdrive/<file id>".

    Examples:
        >>> handler = GoogleDriveHandler()
        >>> handler.get_local_path("gdrive://https://drive.google.com/file/d/14VQf4esuZVy_Xf6IUciBas81j0JpaqCb/view?usp=sharing")  # recommended  # noqa: E501
        >>> # OR
        >>> handler.get_local_path("gdrive://14VQf4esuZVy_Xf6IUciBas81j0JpaqCb")

    Note:
        This handler requires ``gdown`` to be installed. Install it with ``pip install gdown``.
    """

    PREFIX = "gdrive://"

    def _root_dir(self):
        return None

    def _get_local_path(
        self,
        path: str,
        force: bool = False,
        cache_file: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        """Get local path to google drive file.

        To force a download, set ``force=True``.

        Args:
            path (str): The relative file path in the GitHub repository.
                Must start with ``'gdrive://'``.
            force (bool, optional): If ``True``, force a download of the Github
                repository. Defaults to ``False``.
            cache_file (str, optional): The path to cache file to.

        Returns:
            str: The local path to the file/directory.
        """
        if not _GDOWN_AVAILABLE:
            raise ModuleNotFoundError("`gdown` not installed. Install it via `pip install gdown`")
        path = str(path)
        path = path[len(self.PREFIX) :]
        self._check_kwargs(kwargs)

        gdrive_id = None
        if "drive.google.com" in path:
            gdrive_id = path.split("/d/")[1].split("/")[0]
        else:
            gdrive_id = path
        if cache_file is None:
            cache_file = os.path.join(os.path.expanduser("~/.cache/gdrive"), gdrive_id)
        else:
            cache_file = str(cache_file)
        os.makedirs(os.path.dirname(cache_file), exist_ok=True)

        if force or not os.path.exists(cache_file):
            logger = logging.getLogger(__name__)
            logger.info("Downloading gdrive file from {}...".format(path))
            gdown.download(id=gdrive_id, output=cache_file)
            logger.info("File cached to {}".format(cache_file))

        return cache_file


class DownloadHandler(GeneralPathHandler):
    """Handler to download files from a URL.

    Currently this is limited to downloading from Google Drive.

    Examples:
        >>> handler = DownloadHandler()
        >>> handler.get_local_path("
    """

    PREFIX = "download://"

    def __init__(self, path_manager: PathManager, async_executor=None) -> None:
        # Avoid memory leak.
        self.path_manager = weakref.proxy(path_manager)
        super().__init__(async_executor=async_executor)

    def _root_dir(self):
        return None

    def _get_local_path(
        self,
        path: str,
        **kwargs: Any,
    ) -> str:
        """Get local path to google drive file.

        To force a download, set ``force=True``.

        Args:
            path (str): The relative file path in the GitHub repository.
                Must start with ``'download://'``.

        Returns:
            str: The local path to the file/directory.
        """
        path = path[len(self.PREFIX) :]

        if "drive.google.com" in path:
            return self.path_manager.get_local_path(f"gdrive://{path}", **kwargs)
        else:
            raise ValueError(f"Download not supported for url {path}")


def download_github_repository(url, cache_path, branch_or_tag="main", force=False) -> str:
    """Downloads the repository from Github.

    Args:
        version (str): The version to download. Defaults to the
            version of the current codebase.
        path (str): The path to download the repository to.
            Defaults to the the cache directory.
        force (bool): Whether to overwrite existing files.

    Returns:
        str: The path to the downloaded repository.
    """
    url = f"{url}/tarball/{branch_or_tag}"
    cache_path = str(cache_path)
    cache_path = cache_path.format(branch_or_tag=branch_or_tag)

    dir_exists_and_not_empty = os.path.isdir(cache_path) and os.listdir(cache_path)
    if dir_exists_and_not_empty:
        if not force:
            return cache_path
        shutil.rmtree(cache_path)
    os.makedirs(cache_path, exist_ok=True)

    retval = os.system(f"wget -O - {url} | tar xz -C {cache_path}")
    if retval != 0:
        raise RuntimeError("Could not download repository.")

    folders = os.listdir(cache_path)
    assert len(folders) == 1
    curr_path = os.path.join(cache_path, folders[0])
    retval = os.system(f"mv {curr_path}/* {cache_path}")
    if retval != 0:
        raise RuntimeError("Could not download repository.")
    shutil.rmtree(curr_path)

    return cache_path


_path_manager = env.get_path_manager()
_path_manager.register_handler(DataHandler())
_path_manager.register_handler(ResultsHandler())
_path_manager.register_handler(CacheHandler())
_path_manager.register_handler(GithubHandler(env.get_github_url(), default_branch_or_tag=None))
_path_manager.register_handler(AnnotationsHandler())
_path_manager.register_handler(GoogleDriveHandler())
_path_manager.register_handler(DownloadHandler(_path_manager))



import logging

from torch.optim.optimizer import Optimizer

__all__ = ["GradAccumOptimizer"]

logger = logging.getLogger(__name__)


class GradAccumOptimizer(object):
    """Zero grad must be called before step."""

    def __init__(self, optimizer: Optimizer, accumulation_iters: int):
        self.optimizer = optimizer
        self.accumulation_iters = accumulation_iters
        self.step_iters = 0

    def state_dict(self):
        # `step_iters` is not saved because if step_iters > 0,
        # gradients were being accumulated. However when the state dict is saved,
        # there is no way of restoring these gradients.
        # So a `step_iters`>0 would falsely indicate we have some gradients
        # accumulated when we do not.
        return {
            "optimizer": self.optimizer.state_dict(),
            "accumulation_iters": self.accumulation_iters,
        }

    def load_state_dict(self, state_dict):
        if isinstance(state_dict["accumulation_iters"], int):
            # Needed because of a bug in our saving the state dict
            self.accumulation_iters = state_dict["accumulation_iters"]
        else:
            logger.warning(
                f"`accumulation_iters` is of the wrong type due to saving issue. "
                f"Using initialized value of {self.accumulation_iters}."
            )
        return self.optimizer.load_state_dict(state_dict["optimizer"])

    def zero_grad(self):
        if self.step_iters % self.accumulation_iters == 0:
            self.optimizer.zero_grad()

    def step(self, closure=None):
        self.step_iters += 1
        if self.step_iters % self.accumulation_iters == 0:
            self._step(closure)

    def flush(self, closure=None):
        if self.step_iters > 0:
            self._step(closure)
        self.optimizer.zero_grad()

    def _step(self, closure=None):
        """Internal step method. Resets `step_iters` attribute to 0."""
        self.optimizer.step(closure)
        self.step_iters = 0

    def __getattr__(self, item):
        if hasattr(self.optimizer, item):
            return getattr(self.optimizer, item)



import copy
import itertools
import logging
import os
import time
import uuid

import pandas as pd
import torch
from tqdm import tqdm

import meddlr.ops as oF
import meddlr.utils.comm as comm
from meddlr.data.transforms.transform import build_normalizer
from meddlr.evaluation.scan_evaluator import ScanEvaluator, structure_scans
from meddlr.metrics.build import build_metrics

_DEFAULT_METRICS = (
    "DSC",
    "VOE",
    "CV",
    "DSC_scan",
    "VOE_scan",
    "CV_scan",
    "ASSD_scan",
)


class SemSegEvaluator(ScanEvaluator):
    """
    Evaluate semantic segmentation quality.
    """

    def __init__(
        self,
        dataset_name,
        cfg,
        distributed=False,
        sync_outputs=False,
        aggregate_scans=True,
        output_dir=None,
        group_by_scan: bool = False,
        save_seg: bool = False,
        metrics=None,
        flush_period: bool = None,
        to_cpu=True,
    ):
        """
        Args:
            dataset_name (str): name of the dataset to be evaluated.
            cfg (CfgNode): config instance
            output_dir (str): optional, an output directory to dump all
                results predicted on the dataset. The dump contains two files:

                1. "instance_predictions.pth" a file in torch serialization
                   format that contains all the raw original predictions.
                2. "coco_instances_results.json" a json file in COCO's result
                   format.
            distributed (bool, optional): If `True`, collect results from all
                ranks for evaluation.
                Otherwise, will evaluate the results in the current process.
                If using `DistributedDataParallel`, this should likely be `True`.
            sync_outputs (bool, optional): If `True`, synchronizes all predictions
                before evaluation.
                If `False`, synchronizes metrics before reduction.
                Ignored if `distributed=False`
            aggregate_scans (bool, optional): If `True`, also computes metrics per
                scan under the label `'scan_{metric}'` (e.g. scan_l1).

        Not yet supported:
            running_eval (bool, optional): If `True`, evaluates scans while processing.
                This reduces the memory overhead. Note if distributed=`True`, sync_outputs
                must be `False`.
        """
        self._output_dir = output_dir
        self._dataset_name = dataset_name
        self._cpu_device = torch.device("cpu")
        self._logger = logging.getLogger(__name__)
        self._normalizer = build_normalizer(cfg)
        self._distributed = distributed
        self._sync_outputs = sync_outputs
        self._aggregate_scans = aggregate_scans
        self._group_by_scan = group_by_scan
        self._save_seg = save_seg
        self.to_cpu = to_cpu
        self._is_flushing = False

        self.activation = cfg.MODEL.SEG.ACTIVATION
        self.threshold = 0.5
        # if self._distributed and self._sync_outputs and running_eval:
        #     raise ValueError("running_eval not possible when outputs must be synced")

        self.has_bg = cfg.MODEL.SEG.INCLUDE_BACKGROUND or cfg.MODEL.SEG.ACTIVATION == "softmax"
        if self.has_bg:
            # Add "background" to class names if desired.
            self._class_names = (
                cfg.MODEL.SEG.CLASSES
                if any(x.lower() in ["background", "bg"] for x in cfg.MODEL.SEG.CLASSES)
                else ("bg",) + tuple(cfg.MODEL.SEG.CLASSES)
            )
        else:
            self._class_names = cfg.MODEL.SEG.CLASSES

        self._metric_names = []
        if metrics is not False:
            if not metrics:
                metrics = self.default_metrics()
            self._metric_names = metrics

        if flush_period is None:
            flush_period = cfg.TEST.FLUSH_PERIOD
        if distributed and flush_period != 0:
            raise ValueError("Result flushing is not enabled in distributed mode.")
        self.flush_period = flush_period

        self._remaining_state = None
        self._predictions = []
        self._is_flushing = False

    @classmethod
    def default_metrics(cls):
        """The default metrics processed by this class."""
        # TODO: Add support for ASSD_scan
        return _DEFAULT_METRICS

    def reset(self):
        self._predictions = []
        self._remaining_state = None
        self._is_flushing = False

        metrics = self._metric_names
        slice_metrics = [m for m in metrics if not m.endswith("_scan")]
        scan_metrics = [m[:-5] for m in metrics if m.endswith("_scan")]

        self.slice_metrics = build_metrics(slice_metrics, channel_names=self._class_names)
        self.scan_metrics = build_metrics(
            scan_metrics, fmt="{}_scan", channel_names=self._class_names
        )
        self.slice_metrics.eval()
        self.scan_metrics.eval()

    def structure_scans(self, verbose=True):
        """Structure scans into volumes to be used to evaluation."""
        out = structure_scans(self._predictions, verbose=verbose, dims={1: "slice_id"})
        return out

    def process(self, inputs, outputs):
        """
        Args:
            inputs: the inputs to a recon model (e.g., GeneralizedRCNN).
                Currently this should be an empty dictionary.
            outputs: the outputs of a COCO model. It is a list of dicts with key
                "instances" that contains :class:`Instances`.
        """
        labels: torch.Tensor = self._parse_sem_seg_label(inputs)
        pred: torch.Tensor = self._parse_sem_seg_pred(outputs)

        N = pred.shape[0]
        preds = pred.type(torch.bool)
        targets = labels.type(torch.bool)
        if self.to_cpu:
            preds = preds.to(self._cpu_device, non_blocking=True)
            targets = targets.to(self._cpu_device, non_blocking=True)

        self._predictions.extend(
            [
                {
                    "pred": preds[i],
                    "target": targets[i],
                    "metadata": inputs["metadata"][i] if "metadata" in inputs else {},
                }
                for i in range(N)
            ]
        )

        has_num_examples = self.flush_period > 0 and len(self._predictions) >= self.flush_period
        has_num_scans = self.flush_period < 0 and len(
            {x["metadata"]["scan_id"] for x in self._predictions}
        ) > abs(self.flush_period)
        if has_num_examples or has_num_scans:
            self.flush(enter_prediction_scope=True, skip_last_scan=True)

    def synchronize_predictions(self):
        comm.synchronize()
        self._predictions = comm.gather(self._predictions, dst=0)
        self._predictions = list(itertools.chain(*self._predictions))
        if not comm.is_main_process():
            self._predictions = []

    def evaluate(self):
        # Sync predictions (if applicable)
        if self._distributed and self._sync_outputs:
            self.synchronize_predictions()
            if not self._predictions:
                return {}

        if len(self._predictions) == 0:
            self._logger.warning(f"[{type(self).__name__}] Did not receive valid predictions.")
            return {}

        # TODO: Add support in metrics manager for appending metrics to the same scan_id
        for _idx, pred in enumerate(tqdm(self._predictions, desc="Slice Metrics")):
            if "metadata" in pred:
                scan_id = "/".join(
                    str(pred["metadata"].get(k, uuid.uuid4())) for k in ["scan_id", "slice_id"]
                )
                voxel_spacing = pred["metadata"].get("voxel_spacing", None)
                if voxel_spacing:
                    voxel_spacing = voxel_spacing[-(pred["pred"].ndim - 1) :]
            else:
                scan_id = str(uuid.uuid4())
                voxel_spacing = None
            self.slice_metrics(
                ids=[scan_id],
                preds=pred["pred"].unsqueeze(0),
                targets=pred["target"].unsqueeze(0),
                voxel_spacing=voxel_spacing,
                spacing=voxel_spacing,
            )

        # Compute metrics per scan.
        if self._aggregate_scans:
            scans = self.structure_scans()
            for scan_id, pred in tqdm(scans.items(), desc="Scan Metrics"):
                self.scan_metrics(
                    ids=[scan_id],
                    preds=pred["pred"].unsqueeze(0),
                    targets=pred["target"].unsqueeze(0),
                    category_dim=0,
                    voxel_spacing=pred.get("voxel_spacing", None),
                    spacing=pred.get("voxel_spacing", None),
                )

        pred_vals = self.slice_metrics.to_dict()
        pred_vals.update(self.scan_metrics.to_dict())
        self._results = pred_vals

        if not self._is_flushing:
            self.log_summary()

        # Copy so the caller can do whatever with results
        return copy.deepcopy(self._results)

    def log_summary(self):
        if not comm.is_main_process():
            return
        output_dir = self._output_dir
        self._logger.info(
            "[{}] Slice metrics summary:\n{}".format(
                type(self).__name__, self.slice_metrics.summary()
            )
        )
        self._logger.info(
            "[{}] Scan metrics summary:\n{}".format(
                type(self).__name__, self.scan_metrics.summary()
            )
        )

        if not output_dir:
            return

        dirpath = output_dir
        os.makedirs(dirpath, exist_ok=True)
        test_results_summary_path = os.path.join(dirpath, "results.txt")
        slice_metrics_path = os.path.join(dirpath, "slice_metrics.csv")
        scan_metrics_path = os.path.join(dirpath, "scan_metrics.csv")

        # Write details to test file
        with open(test_results_summary_path, "w+") as f:
            f.write("Results generated on %s\n" % time.strftime("%X %x %Z"))
            # f.write("Weights Loaded: %s\n" % os.path.basename(self._config.TEST_WEIGHT_PATH))

            f.write("--" * 40)
            f.write("\n")
            f.write("Slice Metrics:\n")
            f.write(self.slice_metrics.summary())
            f.write("--" * 40)
            f.write("\n")
            f.write("Scan Metrics:\n")
            f.write(self.scan_metrics.summary())
            f.write("--" * 40)
            f.write("\n")

        df: pd.DataFrame = self.slice_metrics.to_pandas()
        df.to_csv(slice_metrics_path, header=True, index=True)

        df: pd.DataFrame = self.scan_metrics.to_pandas()
        df.to_csv(scan_metrics_path, header=True, index=True)

    def _parse_sem_seg_pred(self, input):
        # TODO: Update these values
        num_classes = 1
        channel_dim = 1

        keys = ["sem_seg_pred", "sem_seg_probs", "sem_seg_logits", "pred", "probs", "logits"]
        existing_keys = [k for k in keys if k in input.keys()]
        if len(existing_keys) == 0:
            raise ValueError(f"`input` must have one of keys: {keys}.\n\tGot: {list(input.keys())}")
        key = existing_keys[0]
        assert input[key].shape[1] == len(self._class_names), input[key].shape

        if key.endswith("pred"):
            pred = input[key]
        elif key.endswith("logits"):
            logits = input[key]
            if self.activation == "sigmoid":
                probs = torch.sigmoid(logits)
                pred = (probs >= self.threshold).type(torch.long)
            elif self.activation == "softmax":
                pred = oF.pred_to_categorical(
                    logits, activation=self.activation, channel_dim=1, threshold=0.5
                )
                pred = oF.to_onehot(pred, len(num_classes) + 1)
        elif key.endswith("probs"):
            probs = input[key]
            if self.activation == "sigmoid":
                pred = probs >= self.threshold
            elif self.activation == "softmax":
                pred = oF.to_onehot(
                    torch.argmax(probs, dim=channel_dim),
                    num_classes=len(self._class_names) + 1,
                )

        return pred.type(torch.uint8)

    def _parse_sem_seg_label(self, output):
        keys = ["sem_seg", "sem_seg_labels", "sem_seg_target", "labels", "target"]
        existing_keys = [k for k in keys if k in output.keys()]
        if len(existing_keys) == 0:
            raise ValueError(
                f"`output` must have one of keys: {keys}.\n\tGot: {list(output.keys())}"
            )
        key = existing_keys[0]
        assert output[key].shape[1] == len(self._class_names), f"Key '{key}': {output[key].shape}"
        return output[key].type(torch.uint8)



import json
import logging
import os
import pprint
import re
import sys
from typing import Dict, Mapping

import numpy as np
import pandas as pd
import torch


def print_csv_format(results):
    """Print metrics for easy copypaste.

    Args:
        results (OrderedDict[dict]): task_name -> {metric -> score}
    """
    assert isinstance(results, Dict), results  # unordered results cannot be properly printed
    logger = logging.getLogger(__name__)
    important_res = [(k, v) for k, v in results.items() if "-" not in k]
    logger.info("copypaste: " + ",".join([k[0] for k in important_res]))
    logger.info("copypaste: " + ",".join(["{0:.4f}".format(k[1]) for k in important_res]))

    # Additional formatting
    logger.info(
        "Metrics (comma delimited): \n{}\n{}".format(
            ",".join([k[0] for k in important_res]),
            ",".join(["{0:.4f}".format(k[1]) for k in important_res]),
        )
    )


def verify_results(cfg, results):
    """
    Args:
        results (OrderedDict[dict]): task_name -> {metric -> score}

    Returns:
        bool: whether the verification succeeds or not
    """
    expected_results = cfg.TEST.EXPECTED_RESULTS
    if not len(expected_results):
        return True

    ok = True
    for task, metric, expected, tolerance in expected_results:
        actual = results[task][metric]
        if not np.isfinite(actual):
            ok = False
        diff = abs(actual - expected)
        if diff > tolerance:
            ok = False

    logger = logging.getLogger(__name__)
    if not ok:
        logger.error("Result verification failed!")
        logger.error("Expected Results: " + str(expected_results))
        logger.error("Actual Results: " + pprint.pformat(results))

        sys.exit(1)
    else:
        logger.info("Results verification passed.")
    return ok


def flatten_results_dict(results, delimiter="/"):
    """
    Expand a hierarchical dict of scalars into a flat dict of scalars.
    If results[k1][k2][k3] = v, the returned dict will have the entry
    {"k1/k2/k3": v}.

    Args:
        results (dict):
    """
    r = {}
    for k, v in results.items():
        if isinstance(v, Mapping):
            v = flatten_results_dict(v)
            for kk, vv in v.items():
                r[k + delimiter + kk] = vv
        else:
            r[k] = v
    return r


# Supported validation metrics and the operation to perform on them.
SUPPORTED_VAL_METRICS = {
    "l1": "min",
    "l2": "min",
    "psnr": "max",
    "ssim": "max",
    "ssim_old": "max",
    "ssim (Wang)": "max",
    "l1_scan": "min",
    "l2_scan": "min",
    "psnr_scan": "max",
    "iteration": "max",  # find the last checkpoint
    "loss": "min",
}


def find_weights(cfg, criterion="", iter_limit=None, file_name_fmt="model_{:07d}.pth", top_k=1):
    """Find the best weights based on a validation criterion/metric.

    Args:
        cfg: The config.
        criterion (str, optional): The criterion that we use to select weights.
            Defaults to ``cfg.MODEL.RECON_LOSS.NAME``.
        iter_limit (int, optional): If specified, all weights will be before
            this iteration. If this value is negative, it is
            interpreted as the epoch limit.
        file_name_fmt (int, optional): The naming format for checkpoint files.
        top_k (int, optional): The number of top weights to keep.

    Returns:
        Tuple: ``k`` filepath(s), selection criterion, and ``k`` criterion value(s).
            If ``k=1``, filepath is a string and value is a float.
    """
    logger = logging.getLogger(__name__)

    # Negative iter_limit is interpreted as epoch limit.
    if iter_limit is not None and iter_limit < 0:
        iter_limit = int(abs(iter_limit) * get_iters_per_epoch_eval(cfg))

    ckpt_period = cfg.SOLVER.CHECKPOINT_PERIOD
    eval_period = cfg.TEST.EVAL_PERIOD
    if (
        ckpt_period * eval_period <= 0  # same sign (i.e. same time scale)
        or abs(eval_period) % abs(ckpt_period) != 0  # checkpoint period is multiple of eval period
    ):
        raise ValueError(
            "Cannot find weights if checkpoint/eval periods "
            "at different time scales or eval period is not"
            "a multiple of checkpoint period."
        )

    if not criterion:
        criterion = cfg.MODEL.RECON_LOSS.NAME.lower()
        operation = "min"  # loss is always minimized
    else:
        operation = SUPPORTED_VAL_METRICS[criterion]

    assert operation in ["min", "max"]
    if criterion != "iteration":
        criterion = "val_{}".format(criterion)

    logger.info("Finding best weights in {} using {}...".format(cfg.OUTPUT_DIR, criterion))

    # Filter metrics to find reporting of real validation metrics.
    # If metric is wrapped (e.g. "mridata_knee_2019_val/val_l1"), that means
    # multiple datasets were validated on.
    # We filter out metrics from datasets that contain the word "test".
    # The criterion from all other datasets are averaged and used as the
    # target criterion.
    metrics_file = os.path.join(cfg.OUTPUT_DIR, "metrics")
    metrics = _metrics_from_x(metrics_file, criterion)

    # Filter out all metrics calculated above iter limit.
    if iter_limit:
        metrics = [x for x in metrics if x[0] < iter_limit]

    last_iter = metrics[-1][0]

    # Retraining does not overwrite the metrics file.
    # We make sure that the metrics correspond only to the most
    # recent run.
    metrics_recent_order = metrics[::-1]
    iterations = [m[0] for m in metrics_recent_order]
    intervals = np.diff(iterations)
    if any(intervals > 0):
        stop_idx = np.argmax(intervals > 0) + 1
        metrics_recent_order = metrics_recent_order[:stop_idx]
        metrics = metrics_recent_order[::-1]

    # Note that resuming can sometimes report metrics for the same
    # iteration. We handle this by taking the most recent metric for the
    # iteration __after__ filtering out old training runs.
    # metrics = {iteration: value for iteration, value in metrics}
    metrics = [(iteration, value) for iteration, value in metrics]

    best_iter_and_values = sorted(metrics, key=lambda x: x[1], reverse=operation == "max")[:top_k]

    all_filepaths = []
    all_values = []
    potential_ckpt_files = os.listdir(cfg.OUTPUT_DIR)
    for best_iter, best_value in best_iter_and_values:
        file_name = file_name_fmt.format(best_iter)

        matched_files = [x for x in potential_ckpt_files if re.match(file_name, x)]
        if len(matched_files) > 1:
            raise ValueError(
                f"Too many potential checkpoint files found for iter={best_iter}, "
                f"criterion={criterion}, value={best_value}:\n\t{matched_files}"
            )
        if len(matched_files) == 0:
            if best_iter == last_iter:
                file_name = "model_final.pth"
            matched_files = [re.match(file_name, x) for x in potential_ckpt_files]
        if len(matched_files) == 0:
            raise ValueError(
                f"Could not find potential checkpoint files for iter={best_iter}, "
                f"criterion={criterion}, value={best_value}."
            )
        file_name = matched_files[0]

        file_path = os.path.join(cfg.OUTPUT_DIR, file_name)

        if not os.path.isfile(file_path):
            raise ValueError("Model for iteration {} does not exist".format(best_iter))

        all_filepaths.append(file_path)
        all_values.append(best_value)

        logger.info("Weights: {} - {}: {:0.4f}".format(file_name, criterion, best_value))

    if top_k == 1:
        return all_filepaths[0], criterion, all_values[0]
    else:
        return all_filepaths, criterion, all_values


def check_consistency(state_dict, model):
    """Verifies that the proper weights were loaded into the model.

    Related to issue that loading weights from checkpoints of a cuda model
    does not properly load when model is on cpu. It may also result in
    warnings for contiguous tensors, but this is not always the case.
    https://github.com/pytorch/pytorch/issues/42300

    Args:
        state_dict (Dict): A model state dict.
        model (nn.Module): A Pytorch model.
    """
    _state_dict = model.state_dict()
    for k in state_dict:
        assert torch.equal(state_dict[k], _state_dict[k]), f"Mismatch values: {k}"


def _metrics_from_x(metrics_file, criterion):
    metrics_file = os.path.splitext(metrics_file)[0]
    if os.path.isfile(f"{metrics_file}.json"):
        return _metrics_from_json(f"{metrics_file}.json", criterion)
    elif os.path.isfile(f"{metrics_file}.csv"):
        return _metrics_from_csv(f"{metrics_file}.csv", criterion)
    else:
        raise ValueError(f"metrics file not found - {metrics_file}")


def _metrics_from_json(metrics_file, criterion):
    metrics = []
    with open(metrics_file, "r") as f:
        metrics = [json.loads(line.strip()) for line in f]
    metrics = [m for m in metrics if criterion in m or any(k.endswith(criterion) for k in m.keys())]
    is_metric_wrapped = criterion not in metrics[0]
    if is_metric_wrapped:
        metrics = [
            (
                int(m["iteration"]),
                np.mean(
                    [m[k] for k in m if k.endswith(criterion) and "test" not in k.split("/")[0]]
                ).item(),
            )
            for m in metrics
        ]
    else:
        metrics = [(m["iteration"], m[criterion]) for m in metrics]
    return metrics


def _metrics_from_csv(metrics_file, criterion):
    def _row_has_criterion(row: pd.Series):
        index = row.index.tolist()
        return criterion in index and not pd.isna(row[criterion])

    metrics = pd.read_csv(metrics_file)
    metrics = [m for _, m in metrics.iterrows() if _row_has_criterion(m)]
    metrics = [(int(m["step"]), m[criterion]) for m in metrics]
    return metrics


def get_iters_per_epoch_eval(cfg) -> int:
    """Get number of iterations per epoch for evaluation purposes.

    This function expects a metrics named ``{cfg.OUTPUT_DIR}/metrics.json``
    that is written during training.

    Args:
        cfg: The config.

    Return:
        int: Number of iterations per epoch.
    """
    exp_path = cfg.OUTPUT_DIR

    eval_period = cfg.TEST.EVAL_PERIOD
    time_scale = cfg.TIME_SCALE
    assert time_scale in ["epoch", "iter"]
    if (time_scale == "epoch" and eval_period < 0) or (time_scale == "iter" and eval_period > 0):
        raise ValueError("Evaluation period is not in # epochs")
    eval_period = abs(eval_period)

    with open(os.path.join(exp_path, "metrics.json"), "r") as f:
        metrics = [json.loads(line.strip()) for line in f.readlines()]
    # Filter to only have evaluation metrics.
    metrics = [m for m in metrics if "eval_time" in m]
    # last eval could be at end of training which doesnt have same
    # spacing of iters_per_epoch - skip it
    iterations = np.diff([m["iteration"] for m in metrics][:-1])
    if len(iterations) == 0:
        raise ValueError("Could not determine iters_per_epoch - too few evaluations")
    if len(iterations) >= 2 and not np.all(iterations[:-1] == iterations[0]):
        raise ValueError("Not all iteration spacings are equal - {iterations}")

    iters_per_epoch = iterations[0] / eval_period
    assert iters_per_epoch == int(iters_per_epoch)
    return int(iters_per_epoch)



import ast
import logging
import re
from collections import defaultdict
from typing import Dict, List, Sequence, Tuple, Union

from torch import nn

from meddlr.modeling.layers.build import get_layer_kind
from meddlr.utils.registry import Registry

META_ARCH_REGISTRY = Registry("META_ARCH")  # noqa F401 isort:skip
META_ARCH_REGISTRY.__doc__ = """
Registry for meta-architectures, i.e. the whole model.

The registered object will be called with `obj(cfg)`
and expected to return a `nn.Module` object.
"""

logger = logging.getLogger(__name__)


def build_model(cfg):
    """
    Build the whole model architecture, defined by ``cfg.MODEL.META_ARCHITECTURE``.
    Note that it does not load any weights from ``cfg``.
    """
    meta_arch = cfg.MODEL.META_ARCHITECTURE
    return META_ARCH_REGISTRY.get(meta_arch)(cfg)


def initialize_model(model: nn.Module, initializers: Union[Dict, Tuple]):
    """Initialize the model.

    This function initializes the model using the initialization method
    specified in ``initializers``.

    ``initializers`` should be a sequence of dicts, where each dict
    defines the layer type (optional), regex pattern of the parameter
    name (optional), or the dict. The dict has the following keys:

        * 'kind' (str, optional): The layer kind to apply this (e.g. 'conv', 'norm')
        * 'patterns' (Tuple[str] | str, optional): The regex patterns of the
                parameters to use initializer on. If not specified, all parameters
                of
        * 'initializers' (Sequence[Callable | str]): The initializers to use on the
                parameters. These should be called as ``initializer(param)``.
                These values should be 1:1 with the values in ``'patterns'``.

    Args:
        model (nn.Model): The model to initialize. Parameters will be fetched
            with ``model.named_parameters()``.
        initializers (Dict[str, Union[str, Callable]] | Tuple[str]):
            See above.
    """
    _kind_kwd = "kind"
    _pattern_kwd = "patterns"
    _init_kwd = "initializers"

    if isinstance(initializers, Dict):
        initializers = [initializers]

    # Backwards compatibility with pattern-only initialization.
    if isinstance(initializers, Sequence) and not any(isinstance(x, Dict) for x in initializers):
        assert len(initializers) % 2 == 0, "Sequence of regex_to_init must be even"
        initializers = [
            {_pattern_kwd: k, _init_kwd: v} for k, v in zip(initializers[::2], initializers[1::2])
        ]

    # Convert string values to python literals.
    initializers = _to_literal(initializers)

    if not all(
        isinstance(x, Dict) or (isinstance(x, Sequence) and len(x) == 2) for x in initializers
    ):
        raise ValueError(
            "All initializers must either be a dict or sequence of 2 elements "
            "(pattern, initializer). Got:\n\t{}".format(initializers)
        )

    initializers: List[Dict] = [
        {_pattern_kwd: x[0], _init_kwd: x[1]} if isinstance(x, Sequence) else x
        for x in initializers
    ]
    matched_patterns = {}
    for init_cfg in initializers:
        pattern = init_cfg.pop(_pattern_kwd, None)
        if pattern is None:
            pattern = (".*",)
        elif not isinstance(pattern, Sequence) or isinstance(pattern, str):
            pattern = (pattern,)

        init_method = init_cfg.pop(_init_kwd)
        if not isinstance(init_method, (list, tuple)):
            init_method = (init_method,)

        if len(pattern) != len(init_method):
            raise ValueError(
                "Got {} pattern(s) but {} initializer(s):\n\t"
                "Patterns: {}\n\tInitializers: {}".format(
                    len(pattern), len(init_method), pattern, init_method
                )
            )

        pattern = tuple(
            (p,) if not isinstance(p, Sequence) or isinstance(p, str) else p for p in pattern
        )

        init_method = [
            getattr(nn.init, v)
            if isinstance(v, str)
            else (getattr(nn.init, v[0]), v[1])
            if isinstance(v, Sequence)
            else v
            for v in init_method
        ]
        init_cfg[_init_kwd] = {k: v for k, v in zip(pattern, init_method)}
        for pattern_val in pattern:
            key = (init_cfg[_kind_kwd], pattern_val) if _kind_kwd in init_cfg else pattern_val
            matched_patterns[key] = False

    model_layers = {}
    if any(_kind_kwd in x for x in initializers):
        model_layers: Dict[str, nn.Module] = _get_model_layers(model, by_kind=True)
    named_parameters = list(model.named_parameters())

    # Initialize parameters
    for init_cfg in initializers:
        layer_kind = ""
        if _kind_kwd in init_cfg:
            layer_kind = init_cfg[_kind_kwd]
            if layer_kind not in model_layers:
                logger.warning(
                    f"Layer kind '{layer_kind}' not found in model {type(model).__name__}"
                )
                continue
            parameters = [
                named_param
                for layer in model_layers[layer_kind]
                for named_param in layer.named_parameters()
            ]
        else:
            parameters = named_parameters

        patterns_to_initializers = init_cfg[_init_kwd]
        for name, param in parameters:
            for patterns, v in patterns_to_initializers.items():
                if any(re.match(k, name) for k in patterns):
                    matched_patterns[(layer_kind, patterns) if layer_kind else patterns] = True
                    if isinstance(v, tuple):
                        v[0](param, **v[1])
                    else:
                        v(param)

    unmatched_patterns = [k for k, v in matched_patterns.items() if not v]
    if unmatched_patterns:
        logger.warning(
            "No matches found for these patterns when initialing model '{}':"
            "\n\t- {}\nCheck the model's named parameters: {}\n".format(
                type(model).__name__,
                "\n\t- ".join(str(x) for x in unmatched_patterns),
                [x[0] for x in model.named_parameters()],
            )
        )


def _get_model_layers(model, by_kind=False):
    """Get all layers of a model.

    A layer is defined as a ``nn.Module`` that does not have
    any children.
    """
    modules = model.modules()
    layers = []
    for module in modules:
        has_children = False
        for _ in module.children():
            has_children = True
            break
        if not has_children:
            layers.append(module)

    if by_kind:
        layers_by_kind = defaultdict(list)
        for layer in layers:
            layers_by_kind[get_layer_kind(type(layer))].append(layer)
        return layers_by_kind
    return layers


def _to_literal(x):
    if isinstance(x, str):
        v = x
        if any(
            x.startswith(c1) and x.endswith(c2) for c1, c2 in [("(", ")"), ("[", "]"), ("{", "}")]
        ):
            try:
                v = ast.literal_eval(v)
            except ValueError:
                pass
        return v
    elif isinstance(x, Dict):
        return {_to_literal(k): _to_literal(v) for k, v in x.items()}
    elif isinstance(x, Sequence):
        return type(x)(_to_literal(v) for v in x)
    else:
        return x



"""Transforms specific to medical image data.
Many transform operations are typically written with the assumption of 2D data.
However, medical images typically have >2 dimensions.
Some transforms implemented in this file are meant to overload transforms in the
`fvcore.transforms` module.
"""
import inspect
from typing import Any, Callable, Iterable, List, Mapping, Optional, Tuple, TypeVar

import numpy as np
import torch

from meddlr.transforms.mixins import DeviceMixin

__all__ = ["Transform", "TransformList"]


class Transform(DeviceMixin):
    """
    Base class for implementations of __deterministic__ transfomations for
    _medical_ image and other data structures. Like the `fvcore.transforms`
    module, there should be a higher-level policy that generates (likely with
    random variations) these transform ops.

    By default, all transforms only handle image and segmentation data types.
    Coordinates, bounding boxes, and polygons are not supported by default.
    However, these methods can be overloaded if generalized methods are
    written for these data types.

    Medical images are seldom in the uint8 format and are not always
    normalized between [0, 1] in the floating point format. Transforms should
    not expect that data is normalized in this format.

    Note, each method may choose to modify the input data in-place for
    efficiency.

    This structure is adapted from the fvcore library.
    """

    def _set_attributes(self, params: Optional[Mapping[str, Any]] = None) -> None:
        """
        Set attributes from the input list of parameters.

        Args:
            params (list): list of parameters.
        """

        if params:
            for k, v in params.items():
                if k != "self" and not k.startswith("_"):
                    setattr(self, k, v)

    @classmethod
    def register_type(cls, data_type: str, func: Optional[Callable] = None):
        """
        Register the given function as a handler that this transform will use
        for a specific data type.

        Args:
            data_type (str): the name of the data type (e.g., box)
            func (callable): takes a transform and a data, returns the
                transformed data.

        Examples:

        .. code-block:: python

            # call it directly
            def func(flip_transform, voxel_data):
                return transformed_voxel_data
            HFlipTransform.register_type("voxel", func)

            # or, use it as a decorator
            @HFlipTransform.register_type("voxel")
            def func(flip_transform, voxel_data):
                return transformed_voxel_data

            # ...
            transform = HFlipTransform(...)
            transform.apply_voxel(voxel_data)  # func will be called
        """
        if func is None:  # the decorator style

            def wrapper(decorated_func):
                assert decorated_func is not None
                cls.register_type(data_type, decorated_func)
                return decorated_func

            return wrapper

        assert callable(
            func
        ), "You can only register a callable to a Transform. Got {} instead.".format(func)
        argspec = inspect.getfullargspec(func)
        assert len(argspec.args) == 2, (
            "You can only register a function that takes two positional "
            "arguments to a Transform! Got a function with spec {}".format(str(argspec))
        )
        setattr(cls, "apply_" + data_type, func)

    def apply_image(self, img: torch.Tensor):
        """
        Apply the transform on an image.

        Args:
            img (ndarray): of shape NxHxWxC, or HxWxC or HxW. The array can be
                of type uint8 in range [0, 255], or floating point in range
                [0, 1] or [0, 255].
        Returns:
            ndarray: image after apply the transformation.
        """
        raise NotImplementedError

    def apply_maps(self, maps: torch.Tensor):
        return self.apply_image(maps)

    def inverse(self) -> "Transform":
        """
        Create a transform that inverts the geometric changes (i.e. change of
        coordinates) of this transform.

        Note that the inverse is meant for geometric changes only.
        The inverse of photometric transforms that do not change coordinates
        is defined to be a no-op, even if they may be invertible.

        Returns:
            Transform:
        """
        raise NotImplementedError

    def __repr__(self):
        attrs = {
            k: v
            for k, v in self.__dict__.items()
            if not k.startswith("_") and not isinstance(v, Callable)
        }
        return "{}({})".format(type(self).__name__, ", ".join(f"{k}={v}" for k, v in attrs.items()))

    def __str__(self):
        return self.__repr__()

    def _eq_attrs(self) -> Tuple[str]:
        raise NotImplementedError

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, type(self)):
            return False
        attrs = self._eq_attrs()
        vals = [getattr(self, a) == getattr(o, a) for a in attrs]
        vals = [
            torch.all(v)
            if isinstance(v, torch.Tensor)
            else np.all(v)
            if isinstance(v, (np.ndarray, list, tuple))
            else all(v)
            if isinstance(v, Iterable)
            else v
            for v in vals
        ]
        return all(vals)


class NoOpTransform(Transform):
    def apply_image(self, image):
        return image

    def inverse(self):
        return NoOpTransform()


_T = TypeVar("_T")


# pyre-ignore-all-errors
class TransformList:
    """
    Maintain a list of transform operations which will be applied in sequence.
    Attributes:
        transforms (list[Transform])
    """

    def __init__(self, transforms: List[Transform], ignore_no_op: bool = True):
        """
        Args:
            transforms (list[Transform]): list of transforms to perform.
        """
        super().__init__()
        for t in transforms:
            assert isinstance(t, Transform), t

        self.ignore_no_op = ignore_no_op
        if ignore_no_op:
            transforms = [t for t in transforms if t not in (None, NoOpTransform)]
        self.transforms = transforms

    def _apply(self, x: _T, meth: str) -> _T:
        """
        Apply the transforms on the input.
        Args:
            x: input to apply the transform operations.
            meth (str): meth.
        Returns:
            x: after apply the transformation.
        """
        for t in self.transforms:
            x = getattr(t, meth)(x)
        return x

    def __getattribute__(self, name: str):
        # use __getattribute__ to win priority over any registered dtypes
        if name.startswith("apply_"):
            return lambda x: self._apply(x, name)
        return super().__getattribute__(name)

    def __add__(self, other: "TransformList") -> "TransformList":
        """
        Args:
            other (TransformList): transformation to add.
        Returns:
            TransformList: list of transforms.
        """
        ignore_no_op = self.ignore_no_op or other.ignore_no_op
        if not other:
            return TransformList(self.transforms, ignore_no_op=ignore_no_op)
        others = other.transforms if isinstance(other, TransformList) else [other]
        return TransformList(self.transforms + others, ignore_no_op=ignore_no_op)

    def __iadd__(self, other: "TransformList") -> "TransformList":
        """
        Args:
            other (TransformList): transformation to add.
        Returns:
            TransformList: list of transforms.
        """
        others = other.transforms if isinstance(other, TransformList) else [other]
        if self.ignore_no_op:
            others = [t for t in others if t not in (None, NoOpTransform)]
        self.transforms.extend(others)
        return self

    def __radd__(self, other: "TransformList") -> "TransformList":
        """
        Args:
            other (TransformList): transformation to add.
        Returns:
            TransformList: list of transforms.
        """
        ignore_no_op = self.ignore_no_op or other.ignore_no_op
        others = other.transforms if isinstance(other, TransformList) else [other]
        return TransformList(others + self.transforms, ignore_no_op=ignore_no_op)

    def __len__(self) -> int:
        """
        Returns:
            Number of transforms contained in the TransformList.
        """
        return len(self.transforms)

    def __getitem__(self, idx) -> Transform:
        return self.transforms[idx]

    def __contains__(self, x):
        return x in self.transforms

    def inverse(self) -> "TransformList":
        """
        Invert each transform in reversed order.
        """
        return TransformList([x.inverse() for x in self.transforms[::-1]])

    def __repr__(self):
        return "{}(\n\t{}\n)".format(
            type(self).__name__, "\n\t".join(t.__repr__() for t in self.transforms)
        )

    def __str__(self):
        return self.__repr__()



"""Metric utilities.

All metrics take a reference scan (ref) and an output/reconstructed scan (x).
Both should be tensors with the last dimension equal to 2 (real/imaginary
channels).
"""
from typing import Sequence

import numpy as np
import scipy as scp
import torch
from skimage.metrics import structural_similarity

import meddlr.metrics.functional as mF
from meddlr.metrics.metric import Metric
from meddlr.ops import complex as cplx
from meddlr.utils.deprecated import deprecated

# Mapping from str to complex function name.
_IM_TYPES_TO_FUNCS = {
    "magnitude": cplx.abs,
    "mag": cplx.abs,
    "abs": cplx.abs,
    "phase": cplx.angle,
    "angle": cplx.angle,
    "real": cplx.real,
    "imag": cplx.imag,
}

__all__ = ["PSNR", "MSE", "NRMSE", "RMSE", "SSIM"]


class PSNR(Metric):
    """Peak signal-to-noise ratio with complex-valued support.

    :math:`PSNR = 20 * log_{10}(\\frac{max(|x_{gt}|)}{||x_{pred} - x_{gt}||_2})`

    This implementation supports complex tensors.
    ``im_type`` controls how the complex tensor should be processed:

        - ``'magnitude'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to magnitude images.
        - ``'phase'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to phase images.
        - ``'real'``: Real components of :math:`x_{pred}` and :math:`x_{gt}` are used.
        - ``'imag'``: Imaginary components of :math:`x_{pred}` and :math:`x_{gt}` are used.

    Attributes:
        im_type (str): The type of the complex image to compute the metric on.
            This only applies to complex tensors.
        channel_names (Sequence[str]): The names of the channels in the input.
    """

    is_differentiable = True
    higher_is_better = True

    def __init__(
        self,
        im_type: str = None,
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        super().__init__(
            channel_names=channel_names,
            units="dB",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

        self.im_type = im_type

    def func(self, preds, targets) -> torch.Tensor:
        return mF.psnr(preds, targets, im_type=self.im_type)


class MSE(Metric):
    """Mean squared error with complex-valued support.

    :math:`MSE = ||x_{pred} - x_{gt}||_2^2`.

    This implementation supports complex tensors.
    ``im_type`` controls how the complex tensor should be processed:

        - ``'magnitude'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to magnitude images.
        - ``'phase'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to phase images.
        - ``'real'``: Real components of :math:`x_{pred}` and :math:`x_{gt}` are used.
        - ``'imag'``: Imaginary components of :math:`x_{pred}` and :math:`x_{gt}` are used.

    Attributes:
        im_type (str): The type of the complex image to compute the metric on.
            This only applies to complex tensors.
        channel_names (Sequence[str]): The names of the channels in the input.
    """

    is_differentiable = True
    higher_is_better = False

    def __init__(
        self,
        im_type: str = None,
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

        self.im_type = im_type

    def func(self, preds, targets) -> torch.Tensor:
        return mF.mse(preds, targets, im_type=self.im_type)


class NRMSE(Metric):
    """Normalized root-mean-squared error with complex-valued support.

    :math:`NRMSE = \\frac{||x_{pred} - x_{gt}||_2}{||x_{gt}||_2}`.

    This implementation supports complex tensors.
    ``im_type`` controls how the complex tensor should be processed:

        - ``'magnitude'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to magnitude images.
        - ``'phase'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to phase images.
        - ``'real'``: Real components of :math:`x_{pred}` and :math:`x_{gt}` are used.
        - ``'imag'``: Imaginary components of :math:`x_{pred}` and :math:`x_{gt}` are used.

    Attributes:
        im_type (str): The type of the complex image to compute the metric on.
            This only applies to complex tensors.
        channel_names (Sequence[str]): The names of the channels in the input.
    """

    is_differentiable = True
    higher_is_better = False

    def __init__(
        self,
        im_type: str = None,
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

        self.im_type = im_type

    def func(self, preds, targets) -> torch.Tensor:
        return mF.nrmse(preds, targets, im_type=self.im_type)


nRMSE = NRMSE


class RMSE(Metric):
    """Root-mean-squared error with complex-valued support.

    :math:`RMSE = ||x_{pred} - x_{gt}||_2`.

    This implementation supports complex tensors.
    ``im_type`` controls how the complex tensor should be processed:

        - ``'magnitude'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to magnitude images.
        - ``'phase'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to phase images.
        - ``'real'``: Real components of :math:`x_{pred}` and :math:`x_{gt}` are used.
        - ``'imag'``: Imaginary components of :math:`x_{pred}` and :math:`x_{gt}` are used.

    Attributes:
        im_type (str): The type of the complex image to compute the metric on.
            This only applies to complex tensors.
        channel_names (Sequence[str]): The names of the channels in the input.
    """

    is_differentiable = True
    higher_is_better = False

    def __init__(
        self,
        im_type: str = None,
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

        self.im_type = im_type

    def func(self, preds, targets) -> torch.Tensor:
        return mF.rmse(preds, targets, im_type=self.im_type)


class SSIM(Metric):
    """Structural similarity index measure with complex-valued support.

    This implementation of pSNR supports complex tensors.
    ``im_type`` controls how the complex tensor should be processed:

        - ``'magnitude'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to magnitude images.
        - ``'phase'``: :math:`x_{pred}` and :math:`x_{gt}` are converted to phase images.
        - ``'real'``: Real components of :math:`x_{pred}` and :math:`x_{gt}` are used.
        - ``'imag'``: Imaginary components of :math:`x_{pred}` and :math:`x_{gt}` are used.

    Attributes:
        method (str): The method to use for computing the SSIM.
            Defaults to ``'wang'``.
        im_type (str): The type of the complex image to compute the metric on.
            This only applies to complex tensors.
        channel_names (Sequence[str]): The names of the channels in the input.
    """

    is_differentiable = True
    higher_is_better = True

    def __init__(
        self,
        method: str = "wang",
        im_type: str = "magnitude",
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )
        self.method = method
        self.im_type = im_type

    def func(self, preds, targets) -> torch.Tensor:
        return mF.ssim(
            preds,
            targets,
            method=self.method,
            im_type=self.im_type,
        )


@deprecated(vremove="0.1.0", replacement="metrics.functional.mse")
def compute_mse(ref: torch.Tensor, x: torch.Tensor, is_batch=False, magnitude=False):
    if cplx.is_complex(ref):
        ref = torch.view_as_real(ref)
        x = torch.view_as_real(x)

    assert ref.shape[-1] == 2
    assert x.shape[-1] == 2
    if magnitude:
        squared_err = torch.abs(cplx.abs(x) - cplx.abs(ref)) ** 2
    else:
        squared_err = cplx.abs(x - ref) ** 2
    shape = (x.shape[0], -1) if is_batch else -1
    return torch.mean(squared_err.view(shape), dim=-1)


def compute_l2(ref: torch.Tensor, x: torch.Tensor, is_batch=False, magnitude=False):
    """
    Args:
        ref (torch.Tensor): The target. Shape (...)x2
        x (torch.Tensor): The prediction. Same shape as `ref`.
    """
    if cplx.is_complex(ref):
        ref = torch.view_as_real(ref)
        x = torch.view_as_real(x)

    assert ref.shape[-1] == 2
    assert x.shape[-1] == 2
    return torch.sqrt(compute_mse(ref, x, is_batch=is_batch, magnitude=magnitude))


@deprecated(vremove="0.1.0", replacement="metrics.functional.psnr")
def compute_psnr(ref: torch.Tensor, x: torch.Tensor, is_batch=False, magnitude=False):
    """Compute peak to signal to noise ratio of magnitude image.

    Args:
        ref (torch.Tensor): The target. Shape (...)x2
        x (torch.Tensor): The prediction. Same shape as `ref`.

    Returns:
        Tensor: Scalar in db
    """
    if cplx.is_complex(ref):
        ref = torch.view_as_real(ref)
        x = torch.view_as_real(x)

    assert ref.shape[-1] == 2
    assert x.shape[-1] == 2
    assert not is_batch, "is_batch not supported"
    l2 = compute_l2(ref, x, magnitude=magnitude, is_batch=False)
    # shape = (x.shape[0], -1) if is_batch else -1
    return 20 * torch.log10(cplx.abs(ref).max() / l2)


@deprecated(vremove="0.1.0", replacement="metrics.functional.nrmse")
def compute_nrmse(ref, x, is_batch=False, magnitude=False):
    """Compute normalized root mean square error.
    The norm of reference is used to normalize the metric.

    Args:
        ref (torch.Tensor): The target. Shape (...)x2
        x (torch.Tensor): The prediction. Same shape as `ref`.
    """
    if cplx.is_complex(ref):
        ref = torch.view_as_real(ref)
        x = torch.view_as_real(x)

    assert ref.shape[-1] == 2
    assert x.shape[-1] == 2
    rmse = compute_l2(ref, x, is_batch=is_batch, magnitude=magnitude)
    shape = (x.shape[0], -1) if is_batch else -1
    norm = torch.sqrt(torch.mean((cplx.abs(ref) ** 2).view(shape), dim=-1))

    return rmse / norm


@deprecated(vremove="0.1.0", replacement="metrics.functional.ssim")
def compute_ssim(
    ref: torch.Tensor,
    x: torch.Tensor,
    multichannel: bool = False,
    data_range=None,
    **kwargs,
):
    """Compute structural similarity index metric. Does not preserve autograd.

    Based on implementation of Wang et. al. [1]_

    The image is first converted to magnitude image and normalized
    before the metric is computed.

    Args:
        ref (torch.Tensor): The target. Shape (...)x2
        x (torch.Tensor): The prediction. Same shape as `ref`.
        multichannel (bool, optional): If `True`, computes ssim for real and
            imaginary channels separately and then averages the two.
        data_range(float, optional): The data range of the input image
        (distance between minimum and maximum possible values). By default,
        this is estimated from the image data-type.

    References:
    .. [1] Wang, Z., Bovik, A. C., Sheikh, H. R., & Simoncelli, E. P.
       (2004). Image quality assessment: From error visibility to
       structural similarity. IEEE Transactions on Image Processing,
       13, 600-612.
       https://ece.uwaterloo.ca/~z70wang/publications/ssim.pdf,
       :DOI:`10.1109/TIP.2003.819861`
    """
    gaussian_weights = kwargs.pop("gaussian_weights", True)
    sigma = kwargs.pop("sigma", 1.5)
    use_sample_covariance = kwargs.pop("use_sample_covariance", False)

    if cplx.is_complex(ref):
        ref = torch.view_as_real(ref)
        x = torch.view_as_real(x)

    assert ref.shape[-1] == 2
    assert x.shape[-1] == 2

    if not multichannel:
        ref = cplx.abs(ref)
        x = cplx.abs(x)

    if not x.is_contiguous():
        x = x.contiguous()
    if not ref.is_contiguous():
        ref = ref.contiguous()

    x = x.squeeze().numpy()
    ref = ref.squeeze().numpy()

    if data_range in ("range", "ref-range"):
        data_range = ref.max() - ref.min()
    elif data_range in ("ref-maxval", "maxval"):
        data_range = ref.max()
    elif data_range == "x-range":
        data_range = x.max() - x.min()
    elif data_range == "x-maxval":
        data_range = x.max()

    return structural_similarity(
        ref,
        x,
        data_range=data_range,
        gaussian_weights=gaussian_weights,
        sigma=sigma,
        use_sample_covariance=use_sample_covariance,
        multichannel=multichannel,
        **kwargs,
    )


def compute_vifp_mscale(
    ref: torch.Tensor,
    x: torch.Tensor,
    sigma_nsq: float = 2.0,
    eps: float = 1e-10,
    im_type: str = None,
):
    """Compute visual information fidelity (VIF) metric.

    This code is adapted from
    https://github.com/aizvorski/video-quality/blob/master/vifp.py

    Args:
        ref (torch.Tensor): The reference image. This can be complex.
        x (torch.Tensor): The target image. This can be complex.
        sigma_nsq (float, optional): The visual noise parameter.
            This may need to be fine-tuned over the dataset of interest.
        eps (float, optional): The threshold below which data is considered to be 0.
        im_type (str, optional): The image type to compute metric on.
            Either ``'magnitude'`` (default) to compute metric on magnitude images
            or ``'phase'`` to compute metric on phase/angle images.

    Returns:
        float: The metric value.

    Note:
        ``im_type`` is only valid if input is complex.
    """
    if cplx.is_complex(ref) or cplx.is_complex_as_real(ref):
        ref = _IM_TYPES_TO_FUNCS[im_type](ref)
    if cplx.is_complex(x) or cplx.is_complex_as_real(x):
        x = _IM_TYPES_TO_FUNCS[im_type](x)

    ref = np.squeeze(ref.numpy())
    x = np.squeeze(x.numpy())

    scale_val = 255.0 / ref.max()
    ref *= scale_val
    x *= scale_val

    num = 0.0
    den = 0.0
    for scale in range(1, 5):

        N = 2 ** (4 - scale + 1) + 1
        sd = N / 5.0

        if scale > 1:
            sl = tuple(slice(None, None, 2) if dim > 1 else slice(None) for dim in ref.shape)
            ref = scp.ndimage.gaussian_filter(ref, sd)
            x = scp.ndimage.gaussian_filter(x, sd)
            ref = ref[sl]
            x = x[sl]

        mu1 = scp.ndimage.gaussian_filter(ref, sd)
        mu2 = scp.ndimage.gaussian_filter(x, sd)
        mu1_sq = mu1 * mu1
        mu2_sq = mu2 * mu2
        mu1_mu2 = mu1 * mu2
        sigma1_sq = scp.ndimage.gaussian_filter(ref * ref, sd) - mu1_sq
        sigma2_sq = scp.ndimage.gaussian_filter(x * x, sd) - mu2_sq
        sigma12 = scp.ndimage.gaussian_filter(ref * x, sd) - mu1_mu2

        sigma1_sq[sigma1_sq < 0] = 0
        sigma2_sq[sigma2_sq < 0] = 0

        g = sigma12 / (sigma1_sq + eps)
        sv_sq = sigma2_sq - g * sigma12

        g[sigma1_sq < eps] = 0
        sv_sq[sigma1_sq < eps] = sigma2_sq[sigma1_sq < eps]
        sigma1_sq[sigma1_sq < eps] = 0

        g[sigma2_sq < eps] = 0
        sv_sq[sigma2_sq < eps] = 0

        sv_sq[g < 0] = sigma2_sq[g < 0]
        g[g < 0] = 0
        sv_sq[sv_sq <= eps] = eps

        num += np.sum(np.log10(1 + g * g * sigma1_sq / (sv_sq + sigma_nsq)))
        den += np.sum(np.log10(1 + sigma1_sq / sigma_nsq))

    vifp = num / den

    return vifp


compute_rmse = compute_l2



from typing import Sequence, Union

import torch

from meddlr.transforms.base.motion import MRIMotionTransform
from meddlr.transforms.build import TRANSFORM_REGISTRY
from meddlr.transforms.transform import NoOpTransform
from meddlr.transforms.transform_gen import TransformGen

__all__ = ["RandomMRIMotion"]


@TRANSFORM_REGISTRY.register()
class RandomMRIMotion(TransformGen):
    """A model that corrupts kspace inputs with motion.
    Motion is a common artifact experienced during the MR imaging forward problem.
    When a patient moves, the recorded (expected) location of the kspace sample is
    different than the actual location where the kspace sample that was acquired.
    This module is responsible for simulating different motion artifacts.
    Args:
        seed (int, optional): The fixed seed.
    Attributes:
        generator (torch.Generator): The generator that should be used for all
            random logic in this class.
    Things to consider:
        1. What other information is relevant for inducing motion corruption?
            This could include:
            - ``traj``: The scan trajectory
            - ``etl``: The echo train length - how many readouts per shot.
            - ``num_shots``: Number of shots.
        2. What would a simple translational motion model look?
    Note:
        We do not store this as a module or else it would be saved to the model
        definition, which we dont want.
    """

    _base_transform = MRIMotionTransform

    def __init__(self, std_devs: Union[float, Sequence[float]], p: float = 0.0):
        if isinstance(std_devs, (float, int)):
            std_devs = (std_devs, std_devs)
        elif len(std_devs) > 2:
            raise ValueError("`motion_range` must have 2 or fewer values")
        super().__init__(params={"std_devs": std_devs}, p=p)

    def get_transform(self, input: torch.Tensor):
        params = self._get_param_values(use_schedulers=True)
        if self._rand() >= params["p"]:
            return NoOpTransform()

        std_dev = self._rand_range(*params["std_devs"])
        gen = self._generator
        if gen is None or gen.device != input.device:
            gen = torch.Generator(device=input.device).manual_seed(int(self._rand() * 1e10))
        return MRIMotionTransform(std_dev=std_dev, generator=gen)



from meddlr.transforms.gen.choice import RandomTransformChoice  # noqa
from meddlr.transforms.gen.mask import RandomKspaceMask  # noqa
from meddlr.transforms.gen.motion import RandomMRIMotion  # noqa
from meddlr.transforms.gen.noise import RandomNoise  # noqa
from meddlr.transforms.gen.spatial import (  # noqa
    RandomAffine,
    RandomFlip,
    RandomRot90,
    RandomTranslation,
)



import datetime
import logging
import time
from collections import OrderedDict
from contextlib import contextmanager
from typing import Mapping, Sequence

import torch

from meddlr.utils.env import get_world_size
from meddlr.utils.logger import log_every_n_seconds


class DatasetEvaluator:
    """
    Base class for a dataset evaluator.

    The function :func:`inference_on_dataset` runs the model over
    all samples in the dataset, and have a DatasetEvaluator to process the
    inputs/outputs.

    This class will accumulate information of the inputs/outputs
    (by :meth:`process`),
    and produce evaluation results in the end (by :meth:`evaluate`).
    """

    def reset(self):
        """
        Preparation for a new round of evaluation.
        Should be called before starting a round of evaluation.
        """
        pass

    def process(self, input, output):
        """
        Process an input/output pair.

        Args:
            input: the input that's used to call the model.
            output: the return value of `model(input)`
        """
        pass

    def evaluate(self):
        """
        Evaluate/summarize the performance, after processing all input/output
        pairs.

        Returns:
            dict:
                A new evaluator class can return a dict of arbitrary format
                as long as the user can process the results.
                In our train_net.py, we expect the following format:

                * key: the name of the task (e.g., bbox)
                * value: a dict of {metric name: score}, e.g.: {"AP50": 80}
        """
        pass


class DatasetEvaluators(DatasetEvaluator):
    def __init__(self, evaluators, as_list=False):
        assert len(evaluators)
        super().__init__()
        if isinstance(evaluators, Mapping):
            evaluators = {k: v for k, v in evaluators.items() if v is not None}
        else:
            evaluators = [x for x in evaluators if x is not None]
        self._evaluators = evaluators
        self._as_list = as_list

    def items(self):
        if isinstance(self._evaluators, Mapping):
            return self._evaluators.items()
        return list(enumerate(self._evaluators))

    def values(self) -> Sequence[DatasetEvaluator]:
        if isinstance(self._evaluators, Mapping):
            return self._evaluators.values()
        else:
            return self._evaluators

    def reset(self):
        for evaluator in self.values():
            evaluator.reset()

    def process(self, input, output):
        for evaluator in self.values():
            evaluator.process(input, output)

    def evaluate(self):
        results = [] if self._as_list else OrderedDict()
        for evaluator in self.values():
            result = evaluator.evaluate()
            if self._as_list:
                if result is not None:
                    results.append(result)
                continue

            if result is not None:
                for k, v in result.items():
                    assert (
                        k not in results
                    ), "Different evaluators produce results " "with the same key {}".format(k)
                    results[k] = v
        return results

    def __getitem__(self, index):
        if isinstance(index, str):
            if not isinstance(self._evaluators, Mapping):
                raise ValueError("Cannot index sequence of evaluators with string key.")
            return self._evaluators[index]
        if isinstance(self._evaluators, Mapping):
            return list(self.values())[index]
        else:
            return self._evaluators[index]

    def __len__(self):
        return len(self._evaluators)


def inference_on_dataset(model, data_loader, evaluator):
    """
    Run model on the data_loader and evaluate the metrics with evaluator.
    The model will be used in eval mode.

    Args:
        model (nn.Module): a module which accepts an object from
            `data_loader` and returns some outputs. It will be temporarily set
            to `eval` mode.

            If you wish to evaluate a model in `training` mode instead, you can
            wrap the given model and override its behavior of `.eval()` and
            `.train()`.
        data_loader: an iterable object with a length.
            The elements it generates will be the inputs to the model.
        evaluator (DatasetEvaluator): the evaluator to run. Use
            :class:`DatasetEvaluators([])` if you only want to benchmark, but
            don't want to do any evaluation.

    Returns:
        The return value of `evaluator.evaluate()`
    """
    num_devices = get_world_size()
    logger = logging.getLogger(__name__)

    total = len(data_loader)  # inference data loader must have a fixed length
    logger.info(f"Start inference on {total} batches")
    evaluator.reset()

    num_warmup = min(5, total - 1)
    start_time = time.perf_counter()
    total_compute_time = 0
    with inference_context(model), torch.no_grad():
        for idx, inputs in enumerate(data_loader):
            if idx == num_warmup:
                start_time = time.perf_counter()
                total_compute_time = 0

            start_compute_time = time.perf_counter()
            outputs = model(inputs)
            total_compute_time += time.perf_counter() - start_compute_time

            evaluator.process(inputs, outputs)

            iters_after_start = idx + 1 - num_warmup * int(idx >= num_warmup)
            seconds_per_img = total_compute_time / iters_after_start
            if idx >= num_warmup * 2 or seconds_per_img > 5:
                total_seconds_per_img = (
                    time.perf_counter() - start_time
                ) / iters_after_start  # noqa
                eta = datetime.timedelta(
                    seconds=int(total_seconds_per_img * (total - idx - 1))
                )  # noqa
                log_every_n_seconds(
                    logging.INFO,
                    "Inference done {}/{}. {:.4f} s / img. ETA={}".format(
                        idx + 1, total, seconds_per_img, str(eta)
                    ),
                    n=5,
                )

    # Measure the time only for this worker (before the synchronization barrier)
    total_time = time.perf_counter() - start_time
    total_time_str = str(datetime.timedelta(seconds=total_time))
    # NOTE this format is parsed by grep
    logger.info(
        "Total inference time: "
        "{} ({:.6f} s / batch on {} devices)".format(
            total_time_str, total_time / (total - num_warmup), num_devices
        )
    )
    total_compute_time_str = str(datetime.timedelta(seconds=int(total_compute_time)))
    logger.info(
        "Total inference pure compute time: "
        "{} ({:.6f} s / batch on {} devices)".format(
            total_compute_time_str,
            total_compute_time / (total - num_warmup),
            num_devices,
        )
    )

    eval_start_time = time.perf_counter()
    results = evaluator.evaluate()
    evaluation_time = time.perf_counter() - eval_start_time
    logger.info(f"Evaluation Time: {evaluation_time:.6f} s")
    # An evaluator may return None when not in main process.
    # Replace it by an empty dict instead to make it easier
    # for downstream code to handle
    if results is None:
        results = {}
    return results


@contextmanager
def inference_context(model):
    """
    A context where the model is temporarily changed to eval mode,
    and restored to previous mode afterwards.

    Args:
        model: a torch Module
    """
    training_mode = model.training
    model.eval()
    yield
    model.train(training_mode)



from typing import Union

from fvcore.common.registry import Registry
from torch import nn

CUSTOM_LAYERS_REGISTRY = Registry("CUSTOM_LAYERS")
CUSTOM_LAYERS_REGISTRY.__doc__ = """
Registry for custom layers.

Use this registry to identify if the layer is not provided by default in torch.nn.
"""

_LAYER_SHORTCUTS = {
    "bn": "batchnorm",
    "dropout1d": "dropout",
}
_PT_LAYERS_LOWERCASE = {
    name.lower(): layer
    for name, layer in nn.__dict__.items()
    if isinstance(layer, type) and issubclass(layer, nn.Module)
}

__all__ = ["get_layer_type", "get_layer_kind"]


def get_layer_type(name: str, dimension: int = None):
    """Returns the layer type based on the name and, in some cases, the dimension."""
    in_name = name
    name = name.lower()

    # Handle some layer names earlier because of naming conventions.
    if name == "dropout" and dimension:
        name += f"{dimension}d"

    if name in _LAYER_SHORTCUTS:
        name = _LAYER_SHORTCUTS[name]

    custom_layer_names = {x.lower(): x for x in CUSTOM_LAYERS_REGISTRY._obj_map}
    if name in CUSTOM_LAYERS_REGISTRY:
        return CUSTOM_LAYERS_REGISTRY.get(name)
    if name in custom_layer_names:
        return CUSTOM_LAYERS_REGISTRY.get(custom_layer_names[name])
    if name in _PT_LAYERS_LOWERCASE:
        return _PT_LAYERS_LOWERCASE.get(name)

    if any(x in name for x in ["norm", "conv"]):
        if not dimension:
            raise ValueError(f"{in_name} requires dimension")

    if dimension:
        name += f"{dimension}d"
    if name in CUSTOM_LAYERS_REGISTRY:
        return CUSTOM_LAYERS_REGISTRY.get(name)
    if name in custom_layer_names:
        return CUSTOM_LAYERS_REGISTRY.get(custom_layer_names[name])
    if name in _PT_LAYERS_LOWERCASE:
        return _PT_LAYERS_LOWERCASE.get(name)

    raise ValueError(f"No layer found for '{in_name}'")


def get_layer_kind(layer_type: Union[type, str]):
    if isinstance(layer_type, str):
        layer_type = get_layer_type(layer_type, dimension=2)
    assert issubclass(layer_type, nn.Module)
    kinds_in_name = ["norm", "conv"]
    name = layer_type.__name__
    lower_name = name.lower()
    for kind in kinds_in_name:
        if kind in lower_name:
            return kind

    if hasattr(nn.modules.activation, name):
        return "act"

    if "dropout" in lower_name:
        return "dropout"

    return "unknown"



import os
from collections import defaultdict
from typing import Dict, List

import h5py
import numpy as np
from torch.utils.data import Dataset

from meddlr.data.transforms.transform import DataTransform

__all__ = ["SliceData"]


class SliceData(Dataset):
    """A PyTorch Dataset class that iterates over 2D slices for MRI reconstruction.

    This dataset handles per-slice data iteration for MRI reconstruction problems.
    This means that slices of K-space, target images, and sensitivity maps (if applicable)
    are returned in ``__getitem__``.

    Because slices are the atomic unit in this dataset, we refer to them as the *examples*.
    This dataset splits a single scan (i.e. a dataset dictionary) into multiple slices
    (called *examples*) along the 0-th axis. ``self.transform`` is used to
    execute transforms (e.g. undersampling, intensity normalization, etc.) on the data.

    This dataset is also useful for organizing and fetching supervised and unsupervised
    examples. This functionality is used with :cls:`AlternatingSampler` and
    :cls:`GroupAlternatingSampler`.

    While this class is designed to be flexible for easy interation with different datasets
    (e.g. fastMRI, mridata 2D/3D FSE, SKM-TEA), there are some nuances in how the data must
    be formatted. You can find more information at in the Prepare Your Own Dataset section
    of the documentation (documentation in progress).

    This class was adapted from https://github.com/facebookresearch/fastMRI.
    """

    # Default mapping for key types.
    _DEFAULT_MAPPING = {"kspace": "kspace", "maps": "maps", "target": "target"}
    _REQUIRED_METADATA = ("file_name", "is_unsupervised", "fixed_acc")

    def __init__(self, dataset_dicts: List[Dict], transform, keys=None, include_metadata=False):
        """
        Args:
            dataset_dicts (List[Dict]): List of dictionaries. Each dictionary
                contains information about a single scan in Meddlr format.
            transform (callable): A callable object that pre-processes the
                raw data into appropriate form. The transform function should
                take 'kspace', 'target', 'attributes', 'filename', and 'slice'
                as inputs. 'target' may be null for test data.
            include_metadata (bool, optional): If `True`, includes scan metadata:
                - "scan_id"
                - "slice_id"
        """
        self.transform = transform

        # Convert dataset dict into slices.
        # Each slice is tuple of (file name, slice id, is_unsupervised)
        self.examples = self._init_examples(dataset_dicts)

        # All examples should have the following keys:
        #   - fname (str): The file name
        #   - is_unsupervised (bool): If `True`, the example should be treated as unsupervised
        #   - fixed_acc (float): The fixed acceleration for unsupervised examples.
        #   - Any other keys required for data loading.
        for idx, example in enumerate(self.examples):
            assert all(k in example for k in self._REQUIRED_METADATA), f"Example {idx}"

        self.mapping = dict(self._DEFAULT_MAPPING)
        if keys:
            self.mapping.update(keys)
        self._include_metadata = include_metadata

    def groups(self, group_by):
        _groups = defaultdict(list)
        for idx, example in enumerate(self.examples):
            try:
                _groups[example[group_by]].append(idx)
            except KeyError:
                raise KeyError(f"Key {group_by} not found. Use one of {example.keys()}")
        return _groups

    def _init_examples(self, dataset_dicts):
        examples = []
        for dd in dataset_dicts:
            file_path = dd["file_name"]
            is_unsupervised = dd.get("_is_unsupervised", False)
            acc = dd.get("_acceleration", None)

            if "kspace_size" in dd:
                num_slices = dd["kspace_size"][0]
                shape = dd["kspace_size"][1:3]
            elif "num_slices" in dd:
                num_slices = dd["num_slices"]
                with h5py.File(file_path, "r") as f:
                    shape = f["kspace"].shape[1:3]
            else:
                with h5py.File(file_path, "r") as f:
                    num_slices = f["kspace"].shape[0]
                    shape = f["kspace"].shape[0]

            examples.extend(
                [
                    {
                        "file_name": file_path,
                        "slice_id": slice_id,
                        "is_unsupervised": is_unsupervised,
                        "fixed_acc": acc,
                        "_metadata": dd.get("_metadata", {}),
                        "inplane_shape": shape,
                    }
                    for slice_id in range(num_slices)
                ]
            )
        return examples

    def _load_data(self, example, idx):
        file_path = example["file_name"]
        slice_id = example["slice_id"]
        with h5py.File(file_path, "r") as data:
            kspace = data[self.mapping["kspace"]][slice_id]
            target = data[self.mapping["target"]][slice_id]
            maps = (
                np.zeros_like(target)
                if self.mapping["target"] == "reconstruction_rss"
                else data[self.mapping["maps"]][slice_id]
            )

        return {"kspace": kspace, "maps": maps, "target": target}

    def __getitem__(self, i):
        example = self.examples[i]
        file_path, slice_id, is_unsupervised, fixed_acc = tuple(
            example[k] for k in ["file_name", "slice_id", "is_unsupervised", "fixed_acc"]
        )

        # TODO: remove this forced check.
        if not is_unsupervised:
            fixed_acc = None

        data = self._load_data(example, i)
        kspace = data["kspace"]
        maps = data["maps"]
        target = data["target"]

        fname = os.path.splitext(os.path.basename(file_path))[0]
        masked_kspace, maps, target, mean, std, norm = self.transform(
            kspace, maps, target, fname, slice_id, is_unsupervised, fixed_acc
        )

        vals = {
            "kspace": masked_kspace,
            "maps": maps,
            "mean": mean,
            "std": std,
            "norm": norm,
            "is_unsupervised": is_unsupervised,
        }
        if (
            hasattr(self.transform, "augmentor")
            and isinstance(self.transform, DataTransform)
            and self.transform.augmentor is not None
        ):
            scheduler_params = self.transform.augmentor.get_tfm_gen_params()
            if len(scheduler_params):
                vals["metrics"] = {"scheduler": scheduler_params}
        if self._include_metadata:
            vals["metadata"] = {"scan_id": fname, "slice_id": slice_id}
        if not is_unsupervised:
            vals["target"] = target
        return vals

    def __len__(self):
        return len(self.examples)

    def get_supervised_idxs(self) -> List[int]:
        """Returns indices corresponding to supervised examples.

        These indices can be used with ``__getitem__`` to fetch supervised examples.

        Returns:
            List[int]: The indices corresponding to supervised examples.
        """
        idxs = [idx for idx, x in enumerate(self.examples) if not x["is_unsupervised"]]
        return idxs

    def get_unsupervised_idxs(self):
        """Returns indices corresponding to unsupervised examples.

        These indices can be used with ``__getitem__`` to fetch unsupervised examples.

        Returns:
            List[int]: The indices corresponding to unsupervised examples.
        """
        supervised_idxs = self.get_supervised_idxs()
        unsupervised_idxs = set(range(len(self))) - set(supervised_idxs)
        return sorted(unsupervised_idxs)



"""Event utilities.

This file was adapted from Facebook's detectron2.
https://github.com/facebookresearch/detectron2
"""
import datetime
import json
import logging
import math
import os
from collections import defaultdict
from contextlib import contextmanager

import torch
from fvcore.common.history_buffer import HistoryBuffer

from meddlr.utils.env import get_path_manager, supports_wandb

try:
    import wandb
except ImportError:
    pass

_CURRENT_STORAGE_STACK = []
_PATH_MANAGER = get_path_manager()


def get_event_storage():
    """
    Returns:
        The :class:`EventStorage` object that's currently being used.
        Throws an error if no :class`EventStorage` is currently enabled.
    """
    assert len(_CURRENT_STORAGE_STACK), (
        "get_event_storage() has to be called inside a " "'with EventStorage(...)' context!"
    )
    return _CURRENT_STORAGE_STACK[-1]


class EventWriter:
    """
    Base class for writers that obtain events from :class:`EventStorage` and
    process them.
    """

    def write(self):
        raise NotImplementedError

    def close(self):
        pass


class JSONWriter(EventWriter):
    """
    Write scalars to a json file.

    It saves scalars as one json per line (instead of a big json) for easy
    parsing.

    Examples parsing such a json file:

    .. code-block:: none

        $ cat metrics.json | jq -s '.[0:2]'
        [
          {
            "data_time": 0.008433341979980469,
            "iteration": 20,
            "loss": 1.9228371381759644,
            "loss_box_reg": 0.050025828182697296,
            "loss_classifier": 0.5316952466964722,
            "loss_mask": 0.7236229181289673,
            "loss_rpn_box": 0.0856662318110466,
            "loss_rpn_cls": 0.48198649287223816,
            "lr": 0.007173333333333333,
            "time": 0.25401854515075684
          },
          {
            "data_time": 0.007216215133666992,
            "iteration": 40,
            "loss": 1.282649278640747,
            "loss_box_reg": 0.06222952902317047,
            "loss_classifier": 0.30682939291000366,
            "loss_mask": 0.6970193982124329,
            "loss_rpn_box": 0.038663312792778015,
            "loss_rpn_cls": 0.1471673548221588,
            "lr": 0.007706666666666667,
            "time": 0.2490077018737793
          }
        ]

        $ cat metrics.json | jq '.loss_mask'
        0.7126231789588928
        0.689423680305481
        0.6776131987571716
        ...

    """

    def __init__(self, json_file, window_size=20):
        """
        Args:
            json_file (str): path to the json file. New data will be appended
                if the file exists.
            window_size (int): the window size of median smoothing for the
                scalars whose `smoothing_hint` are True.
        """
        self._file_handle = _PATH_MANAGER.open(json_file, "a")
        self._window_size = window_size

    def write(self):
        storage = get_event_storage()
        to_save = {"iteration": storage.iter}
        to_save.update(storage.latest_with_smoothing_hint(self._window_size))
        self._file_handle.write(json.dumps(to_save, sort_keys=True) + "\n")
        self._file_handle.flush()
        try:
            os.fsync(self._file_handle.fileno())
        except AttributeError:
            pass

    def close(self):
        self._file_handle.close()


class TensorboardXWriter(EventWriter):
    """
    Write all scalars to a tensorboard file.
    """

    def __init__(self, log_dir: str, window_size: int = 20, **kwargs):
        """
        Args:
            log_dir (str): the directory to save the output events
            window_size (int): the scalars will be median-smoothed by this
                window size

            kwargs: other arguments passed to
                `torch.utils.tensorboard.SummaryWriter(...)`
        """
        self._window_size = window_size
        from torch.utils.tensorboard import SummaryWriter

        self._writer = SummaryWriter(log_dir, **kwargs)

    def write(self):
        storage = get_event_storage()
        for k, v in storage.latest_with_smoothing_hint(self._window_size).items():
            self._writer.add_scalar(k, v, storage.iter)

        if len(storage.vis_data) >= 1:
            for img_name, img, step_num, data_format in storage.vis_data:
                self._writer.add_image(img_name, img, step_num, dataformats=data_format)
            storage.clear_images()

    def close(self):
        if hasattr(self, "_writer"):  # doesn't exist when the code fails at import
            self._writer.close()


class CommonMetricPrinter(EventWriter):
    """
    Print **common** metrics to the terminal, including
    iteration time, ETA, memory, all losses, and the learning rate.

    To print something different, please implement a similar printer by
    yourself.
    """

    def __init__(self, max_iter, eval_period=0):
        """
        Args:
            max_iter (int): the maximum number of iterations to train.
                Used to compute ETA.
            eval_period (int): the period for performing evaluation.
                Specifying this will result in a better ETA estimate.
                Assumes evaluation will be done at least once when training ends
                and key "eval_time" recorded in `storage.history`.
        """
        self.logger = logging.getLogger(__name__)
        self._max_iter = max_iter
        self._eval_period = eval_period

    def write(self):
        storage = get_event_storage()
        iteration = storage.iter

        data_time, time = None, None
        eta_string = "N/A"
        try:
            data_time = storage.history("data_time").avg(20)
            time = storage.history("time").global_avg()
            eta_seconds = storage.history("time").median(1000) * (self._max_iter - iteration)
            if self._eval_period and "eval_time" in storage.histories():
                num_eval_done = (iteration + 1) // self._eval_period
                num_eval_remaining = math.ceil(self._max_iter / self._eval_period) - num_eval_done
                eta_seconds += storage.history("eval_time").avg(2) * num_eval_remaining

            storage.put_scalar("eta_seconds", eta_seconds, smoothing_hint=False)
            eta_string = str(datetime.timedelta(seconds=int(eta_seconds)))
        except KeyError:  # they may not exist in the first few iterations (due to warmup)  # noqa: E501
            pass

        try:
            lr = "{:.6f}".format(storage.history("lr").latest())
        except KeyError:
            lr = "N/A"

        if torch.cuda.is_available():
            max_mem_mb = torch.cuda.max_memory_allocated() / 1024.0 / 1024.0
        else:
            max_mem_mb = None

        ram_key = "guppy/total_RAM"
        if ram_key in storage.histories():
            total_ram = storage.history(ram_key).latest()  # this value is already in MB
        else:
            total_ram = None

        # NOTE: max_mem is parsed by grep in "dev/parse_results.sh"
        self.logger.info(
            """\
eta: {eta}  iter: {iter}  {losses}  \
{time}  {data_time}  \
lr: {lr}  {memory}  {ram}\
""".format(
                eta=eta_string,
                iter=iteration,
                losses="  ".join(
                    [
                        ("{}: {:.3%s}" % ("f" if v.median(20) >= 0.0005 else "e")).format(
                            k, v.median(20)
                        )
                        for k, v in storage.histories().items()
                        if "loss" in k
                    ]
                ),
                time="time: {:.4f}".format(time) if time is not None else "",
                data_time="data_time: {:.4f}".format(data_time) if data_time is not None else "",
                lr=lr,
                memory="max_mem: {:.0f}M".format(max_mem_mb) if max_mem_mb is not None else "",
                ram="total_RAM: {:.0f}M".format(total_ram) if total_ram is not None else "",
            )
        )


class WandBWriter(EventWriter):
    """Write scalars with Weights and Biases.

    If `wandb.init` is called with `sync_tensorboard=True`, this writer may not be needed.
    When synced with tensorboard, W&B automatically scrapes tensorboard for logged data.
    Therefore, you would only need to log to tensorboard and it would be copied over.

    Note there is a bug related to tensorboard step/iteration count not corresponding with
    W&B steps/iteration count
    """

    def __init__(self, window_size=20):
        """
        Args:
            window_size (int): the window size of median smoothing for the
                scalars whose `smoothing_hint` are True.
        """
        if not supports_wandb():
            raise ModuleNotFoundError("Module `wandb` not found or running in debug mode.")
        if wandb.run is None:
            raise ValueError("wand.run not found. Call `wand.init` before creating writer.")
        self._window_size = window_size

    def write(self):
        storage = get_event_storage()
        to_save = storage.latest_with_smoothing_hint(self._window_size)
        wandb.log(to_save, step=storage.iter)

        # TODO: Add logging images.


class EventStorage:
    """
    The user-facing class that provides metric storage functionalities.

    In the future we may add support for storing / logging other types of data
    if needed.
    """

    def __init__(self, start_iter=0):
        """
        Args:
            start_iter (int): the iteration number to start with
        """
        self._history = defaultdict(HistoryBuffer)
        self._smoothing_hints = {}
        self._latest_scalars = {}
        self._iter = start_iter
        self._current_prefix = ""
        self._vis_data = []

    def put_image(self, img_name, img_tensor, data_format="HWC"):
        """
        Add an `img_tensor` to the `_vis_data` associated with `img_name`.

        Args:
            img_name (str): The name of the image to put into tensorboard.
            img_tensor (torch.Tensor or numpy.array): An `uint8` or `float`
                Tensor of shape `[channel, height, width]` where `channel` is
                3. The image format should be RGB. The elements in img_tensor
                can either have values in [0, 1] (float32) or [0, 255] (uint8).
                The `img_tensor` will be visualized in tensorboard.
            data_format (str): Either "CHW", "HW", or "HWC"
        """
        self._vis_data.append((img_name, img_tensor, self._iter, data_format))

    def clear_images(self):
        """
        Delete all the stored images for visualization. This should be called
        after images are written to tensorboard.
        """
        self._vis_data = []

    def put_scalar(self, name, value, smoothing_hint=True):
        """
        Add a scalar `value` to the `HistoryBuffer` associated with `name`.

        Args:
            smoothing_hint (bool): a 'hint' on whether this scalar is noisy and
                should be
                smoothed when logged. The hint will be accessible through
                :meth:`EventStorage.smoothing_hints`.  A writer may ignore the
                hint and apply custom smoothing rule.

                It defaults to True because most scalars we save need to be
                smoothed to
                provide any useful signal.
        """
        name = self._current_prefix + name
        history = self._history[name]
        value = float(value)
        history.update(value, self._iter)
        self._latest_scalars[name] = value

        existing_hint = self._smoothing_hints.get(name)
        if existing_hint is not None:
            assert (
                existing_hint == smoothing_hint
            ), "Scalar {} was put with a different smoothing_hint!".format(name)
        else:
            self._smoothing_hints[name] = smoothing_hint

    def put_scalars(self, *, smoothing_hint=True, **kwargs):
        """
        Put multiple scalars from keyword arguments.

        Examples:

            storage.put_scalars(
                loss=my_loss, accuracy=my_accuracy, smoothing_hint=True
            )
        """
        for k, v in kwargs.items():
            self.put_scalar(k, v, smoothing_hint=smoothing_hint)

    def history(self, name):
        """
        Returns:
            HistoryBuffer: the scalar history for name
        """
        ret = self._history.get(name, None)
        if ret is None:
            raise KeyError("No history metric available for {}!".format(name))
        return ret

    def histories(self):
        """
        Returns:
            dict[name -> HistoryBuffer]: the HistoryBuffer for all scalars
        """
        return self._history

    def latest(self):
        """
        Returns:
            dict[name -> number]: the scalars that's added in the current
                iteration.
        """
        return self._latest_scalars

    def latest_with_smoothing_hint(self, window_size=20):
        """
        Similar to :meth:`latest`, but the returned values
        are either the un-smoothed original latest value,
        or a median of the given window_size,
        depend on whether the smoothing_hint is True.

        This provides a default behavior that other writers can use.
        """
        result = {}
        for k, v in self._latest_scalars.items():
            result[k] = self._history[k].median(window_size) if self._smoothing_hints[k] else v
        return result

    def smoothing_hints(self):
        """
        Returns:
            dict[name -> bool]: the user-provided hint on whether the scalar
                is noisy and needs smoothing.
        """
        return self._smoothing_hints

    def step(self):
        """
        User should call this function at the beginning of each iteration, to
        notify the storage of the start of a new iteration.
        The storage will then be able to associate the new data with the
        correct iteration number.
        """
        self._iter += 1
        self._latest_scalars = {}

    @property
    def vis_data(self):
        return self._vis_data

    @property
    def iter(self):
        return self._iter

    @property
    def iteration(self):
        # for backward compatibility
        return self._iter

    def __enter__(self):
        _CURRENT_STORAGE_STACK.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert _CURRENT_STORAGE_STACK[-1] == self
        _CURRENT_STORAGE_STACK.pop()

    @contextmanager
    def name_scope(self, name):
        """
        Yields:
            A context within which all the events added to this storage
            will be prefixed by the name scope.
        """
        old_prefix = self._current_prefix
        self._current_prefix = name.rstrip("/") + "/"
        yield
        self._current_prefix = old_prefix



import inspect
import itertools
from typing import Any, Dict, Sequence

import pandas as pd
import torch
from torchmetrics.metric import Metric as _Metric
from torchmetrics.utilities import reduce

from meddlr.utils import comm

__all__ = ["Metric"]


class Metric(_Metric):
    """Interface for new metrics.

    A metric should be implemented as a callable with explicitly defined
    arguments. In other words, metrics should not have `**kwargs` or `**args`
    options in the `__call__` method.

    While not explicitly constrained to the return type, metrics typically
    return float value(s). The number of values returned corresponds to the
    number of categories.

    This class is opinionated in that it computes metrics for each (example, channel)
    pair. This means that outputs of ``compute`` are not scalars, but rather tensors
    of shape ``(B, C)``. Note, this opinion may be relaxed in the future.

    * metrics should have different name() for different functionality.
    * `category_dim` duck type if metric can process multiple categories at
        once.

    To compute metrics:

    .. code-block:: python

        metric = Metric()
        results = metric(...)
    """

    def __init__(
        self,
        channel_names: Sequence[str] = None,
        units: str = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        self.units = units
        super().__init__(
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

        self.reduction = reduction
        self.channel_names = channel_names
        self._update_kwargs_aliases = {}

        # Identifiers for the examples that are seen.
        self.add_state("ids", default=[], dist_reduce_fx=lambda x: itertools.chain(*x))
        self.add_state("values", default=[], dist_reduce_fx="cat")

    def func(self, preds, targets, *args, **kwargs) -> torch.Tensor:
        """Computes metrics for each element in the batch.

        Returns:
            torch.Tensor: A torch Tensor with first dimension being
                batch dimension (``Bx...``).
        """
        raise NotImplementedError

    def update(self, preds, targets, *args, ids=None, **kwargs):
        assert preds.shape == targets.shape

        values: torch.Tensor = self.func(preds, targets, *args, **kwargs)
        self.values.append(values)
        self._add_ids(ids=ids, num_samples=len(values))

    def _generate_ids(self, num_samples):
        id_start = sum(len(x) for x in self.values)
        rank = comm.get_rank()
        ids = [f"{rank}-{id_start + idx}" for idx in range(num_samples)]
        return ids

    def _add_ids(self, ids, num_samples):
        if ids is None:
            ids = self._generate_ids(num_samples)
        self.ids.extend(ids)

    def compute(self, reduction=None):
        if reduction is None:
            reduction = self.reduction
        return reduce(torch.cat(self.values), reduction)

    def to_pandas(self, sync_dist: bool = True) -> pd.DataFrame:
        return pd.DataFrame.from_dict(self.to_dict(sync_dist=sync_dist, device="cpu"))

    def to_dict(self, sync_dist: bool = True, device=None):
        if sync_dist:
            with self.sync_context():
                data = self._to_dict(device=device)
        else:
            data = self._to_dict(device=device)
        return data

    def _to_dict(self, device=None):
        if not self.values:
            return {"id": self.ids}

        values = torch.cat(self.values) if isinstance(self.values, list) else self.values
        if device is not None:
            values = values.to(device)
        channel_names = (
            self.channel_names
            if self.channel_names
            else [f"channel_{idx}" for idx in range(values.shape[1])]
        )
        data = {"id": self.ids}
        data.update({name: values[:, idx] for idx, name in enumerate(channel_names)})
        return data

    def _filter_kwargs(self, **kwargs: Any) -> Dict[str, Any]:
        """filter kwargs such that they match the update signature of the metric"""
        if self._update_kwargs_aliases:
            filtered_kwargs = {k: v for k, v in kwargs.items()}
            aliases = {}
            for alias in self._update_kwargs_aliases:
                if alias not in kwargs or self._update_kwargs_aliases[alias] in aliases:
                    continue
                aliases[self._update_kwargs_aliases[alias]] = kwargs.pop(alias)
            filtered_kwargs.update(aliases)
        else:
            filtered_kwargs = kwargs
        return super()._filter_kwargs(**filtered_kwargs)

    def register_update_aliases(self, **kwargs):
        """Register aliases for keyword arguments when calling update."""
        # filter all parameters based on update signature except those of
        # type VAR_POSITIONAL (*args) and VAR_KEYWORD (**kwargs)
        _params = (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
        _sign_params = self._update_signature.parameters
        supported_kwargs = tuple(
            k for k in _sign_params.keys() if _sign_params[k].kind not in _params
        )
        unsupported_kwargs = [v for v in kwargs.values() if v not in supported_kwargs]
        if len(unsupported_kwargs) > 0:
            raise ValueError(
                f"Found unsupported kwargs '{unsupported_kwargs}'. "
                f"Supported keyword arguments include:{supported_kwargs}"
            )
        aliases = {k: v for k, v in kwargs.items()}
        self._update_kwargs_aliases.update(aliases)

    def name(self):
        return type(self).__name__

    def display_name(self):
        """Name to use for pretty printing and display purposes."""
        name = self.name()
        return "{} ({})".format(name, self.units) if self.units else name



import copy
import inspect
import logging
import types
from collections import defaultdict
from typing import List

from tabulate import tabulate

from meddlr.utils.logger import log_first_n

__all__ = ["DatasetCatalog", "MetadataCatalog"]


class DatasetCatalog(object):
    """A catalog that stores information about the datasets and how to fetch them.

    This catalog makes it easy to choose different datasets based on a config string.

    It contains a mapping from strings
    (which are names that identify a dataset, e.g. "mridata_knee_2019_train")
    to a function which parses the dataset and returns the samples in the
    format of `list[dict]`.

    The returned dicts should be in Meddlr Dataset format
    (See datasets/README.md for details). For a list of built-in datasets,
    see `data/datasets/builtin.py`.

    This catalog also supports aliases for different datasets. This is useful for
    backwards compatibility and easy config access.

    Adapted from https://github.com/facebookresearch/detectron2.

    TODO (arjundd): Make this based off Registry.
    """

    _REGISTERED = {}
    _ALIASES = {}

    @staticmethod
    def register(name, func_or_name):
        """
        Args:
            name (str): The name that identifies a dataset,
                e.g. "mridata_knee_2019_train".
            func_or_name (callable | str): A callable that returns a list of dicts
                or the name of an existing dataset. If the latter, it will be registered
                as a soft alias to the underlying dataset.
        """
        if name in DatasetCatalog._REGISTERED:
            raise ValueError("Dataset '{}' is already registered!".format(name))
        if isinstance(func_or_name, str):
            assert (
                func_or_name in DatasetCatalog._REGISTERED
            ), f"Target dataset '{func_or_name}' is not registered"
            DatasetCatalog._ALIASES[name] = func_or_name
            return

        assert callable(
            func_or_name
        ), "You must register a function with `DatasetCatalog.register`!"
        DatasetCatalog._REGISTERED[name] = (func_or_name, inspect.signature(func_or_name))

    @staticmethod
    def get(name, *args, **kwargs):
        """
        Call the registered function and return its results.

        Args:
            name (str): the name that identifies a dataset,
                e.g. "coco_2014_train".

        Returns:
            list[dict]: dataset annotations.0
        """
        try:
            dname = DatasetCatalog._ALIASES.get(name, name)
            f, sig = DatasetCatalog._REGISTERED[dname]
        except KeyError:
            raise KeyError(
                "Dataset '{}' is not registered! "
                "Available datasets are: {}".format(
                    name, ", ".join(DatasetCatalog._REGISTERED.keys())
                )
            )

        logger = logging.getLogger(__name__)
        parameters = sig.parameters
        has_kwargs = any(v.kind == inspect.Parameter.VAR_KEYWORD for v in parameters.values())
        if not has_kwargs:
            missing_kwargs = [k for k in kwargs if k not in parameters]
            valid_kwargs = {k: v for k, v in kwargs.items() if k in parameters}
            if missing_kwargs:
                logger.warning(
                    "Function to load dataset {} does not support following kwargs. "
                    "Ignoring them...\n\t{}".format(name, missing_kwargs)
                )
            kwargs = valid_kwargs
        return f(*args, **kwargs)

    @staticmethod
    def list() -> List[str]:
        """List all registered datasets (including aliases).

        For a more decorated output, see :func:`DatasetCatalog.__repr__`.

        Returns:
            list[str]
        """
        return list(DatasetCatalog._REGISTERED.keys()) + list(DatasetCatalog._ALIASES.keys())

    @staticmethod
    def clear(name=None):
        """Remove the registered dataset and it's aliases.

        Args:
            name (str): The name of the dataset to be removed.
                If None, all datasets are removed.
        """
        if name is None:
            DatasetCatalog._REGISTERED.clear()
            DatasetCatalog._ALIASES.clear()
            return

        aliases = [name] + [k for k, v in DatasetCatalog._ALIASES.items() if v == name]
        DatasetCatalog._REGISTERED.pop(name, None)
        for alias in aliases:
            DatasetCatalog._ALIASES.pop(alias, None)

    def __repr__(self) -> str:
        base_to_aliases = defaultdict(list)
        for alias, base in self._ALIASES.items():
            base_to_aliases[base].append(alias)

        datasets = [
            {
                "name": name,
                "aliases": base_to_aliases.get(name, None),
                **MetadataCatalog.get(name, {}).__dict__,  # use protected variable to avoid a copy
            }
            for name in self._REGISTERED
        ]
        table = tabulate(datasets, headers="keys", tablefmt="fancy_grid")
        return "Cataloged Datasets:\n{}".format(table)


class Metadata(types.SimpleNamespace):
    """
    A class that supports simple attribute setter/getter.
    It is intended for storing metadata of a dataset and make it accessible
    globally.

    Examples:

    .. code-block:: python

        # somewhere when you load the data:
        MetadataCatalog.get("mydataset").thing_classes = ["person", "dog"]

        # somewhere when you print statistics or visualize:
        classes = MetadataCatalog.get("mydataset").thing_classes
    """

    # the name of the dataset
    # set default to N/A so that `self.name` in the errors
    # will not trigger getattr again
    name: str = "N/A"

    _RENAMED = {
        "class_names": "thing_classes",
        "dataset_id_to_contiguous_id": "thing_dataset_id_to_contiguous_id",
        "stuff_class_names": "stuff_classes",
    }

    def __getattr__(self, key):
        if key in self._RENAMED:
            log_first_n(
                logging.WARNING,
                "Metadata '{}' was renamed to '{}'!".format(key, self._RENAMED[key]),
                n=10,
            )
            return getattr(self, self._RENAMED[key])

        raise AttributeError(
            "Attribute '{}' does not exist in the metadata of '{}'. "
            "Available keys are {}.".format(key, self.name, str(self.__dict__.keys()))
        )

    def __setattr__(self, key, val):
        if key in self._RENAMED:
            log_first_n(
                logging.WARNING,
                "Metadata '{}' was renamed to '{}'!".format(key, self._RENAMED[key]),
                n=10,
            )
            setattr(self, self._RENAMED[key], val)

        # Ensure that metadata of the same name stays consistent
        try:
            oldval = getattr(self, key)
            assert oldval == val, (
                "Attribute '{}' in the metadata of '{}' cannot be set "
                "to a different value!\n{} != {}".format(key, self.name, oldval, val)
            )
        except AttributeError:
            super().__setattr__(key, val)

    def as_dict(self):
        """
        Returns all the metadata as a dict.
        Note that modifications to the returned dict will not reflect on the
        Metadata object.
        """
        return copy.copy(self.__dict__)

    def set(self, **kwargs):
        """
        Set multiple metadata with kwargs.
        """
        for k, v in kwargs.items():
            setattr(self, k, v)
        return self

    def get(self, key, default=None):
        """
        Access an attribute and return its value if exists.
        Otherwise return default.
        """
        try:
            return getattr(self, key)
        except AttributeError:
            return default


class MetadataCatalog:
    """
    MetadataCatalog provides access to "Metadata" of a given dataset.

    The metadata associated with a certain name is a singleton: once created,
    the metadata will stay alive and will be returned by future calls to
    `get(name)`.

    It's like global variables, so don't abuse it.
    It's meant for storing knowledge that's constant and shared across the
    execution of the program, e.g.: the class names in COCO.

    Adapted from https://github.com/facebookresearch/detectron2.
    """

    _NAME_TO_META = {}

    @staticmethod
    def get(name):
        """
        Args:
            name (str): name of a dataset (e.g. coco_2014_train).

        Returns:
            Metadata: The :class:`Metadata` instance associated with this name,
            or create an empty one if none is available.
        """
        assert len(name)
        if name in MetadataCatalog._NAME_TO_META:
            ret = MetadataCatalog._NAME_TO_META[name]
            return ret
        else:
            m = MetadataCatalog._NAME_TO_META[name] = Metadata(name=name)
            return m



from meddlr.transforms.base import *  # noqa: F401, F403
from meddlr.transforms.gen import *  # noqa: F401, F403
from meddlr.transforms.transform import Transform, TransformList  # noqa: F401
from meddlr.transforms.transform_gen import TransformGen  # noqa: F401
